"""
Airspeed Velocity benchmark utilities
"""
from __future__ import division, absolute_import, print_function

import sys
import re
import time
import textwrap
import subprocess


class Benchmark(object):
    """
    Base class with sensible options
    """
    pass


def run_monitored(code):
    """
    Run code in a new Python process, and monitor peak memory usage.

    Returns
    -------
    duration : float
        Duration in seconds (including Python startup time)
    peak_memusage : float
        Peak memory usage (rough estimate only) in bytes

    """
    if not sys.platform.startswith('linux'):
        raise RuntimeError("Peak memory monitoring only works on Linux")

    code = textwrap.dedent(code)
    process = subprocess.Popen([sys.executable, '-c', code])

    peak_memusage = -1

    start = time.time()
    while True:
        ret = process.poll()
        if ret is not None:
            break

        with open('/proc/%d/status' % process.pid, 'r') as f:
            procdata = f.read()

        m = re.search('VmRSS:\s*(\d+)\s*kB', procdata, re.S | re.I)
        if m is not None:
            memusage = float(m.group(1)) * 1e3
            peak_memusage = max(memusage, peak_memusage)

        time.sleep(0.01)

    process.wait()

    duration = time.time() - start

    if process.returncode != 0:
        raise AssertionError("Running failed:\n%s" % code)

    return duration, peak_memusage


def get_mem_info():
    """Get information about available memory"""
    if not sys.platform.startswith('linux'):
        raise RuntimeError("Memory information implemented only for Linux")

    info = {}
    with open('/proc/meminfo', 'r') as f:
        for line in f:
            p = line.split()
            info[p[0].strip(':').lower()] = float(p[1]) * 1e3
    return info


def set_mem_rlimit(max_mem=None):
    """
    Set address space rlimit
    """
    import resource
    if max_mem is None:
        mem_info = get_mem_info()
        max_mem = int(mem_info['memtotal'] * 0.7)
    cur_limit = resource.getrlimit(resource.RLIMIT_AS)
    if cur_limit[0] > 0:
        max_mem = min(max_mem, cur_limit[0])

    resource.setrlimit(resource.RLIMIT_AS, (max_mem, cur_limit[1]))


def with_attributes(**attrs):
    def decorator(func):
        for key, value in attrs.items():
            setattr(func, key, value)
        return func
    return decorator



# -*- coding: utf-8 -*-

from __future__ import division, print_function

from .move import Move

from .mh import MHMove
from .gaussian import GaussianMove

from .red_blue import RedBlueMove
from .stretch import StretchMove
from .walk import WalkMove
from .kde import KDEMove

from .de import DEMove
from .de_snooker import DESnookerMove

__all__ = [
    "Move",
    "MHMove", "GaussianMove",
    "RedBlueMove", "StretchMove", "WalkMove", "KDEMove",
    "DEMove", "DESnookerMove",
]



from refnx.util.ErrorProp import (EPadd, EPsub, EPdiv, EPmul, EPcos, EPexp,
                                  EPlog, EPlog10, EPmulk, EPpow, EPpowk, EPsin,
                                  EPtan)
from refnx.util.nsplice import get_scaling_in_overlap
from refnx.util.general import (div, q, q2, qcrit, double_chopper_frequency,
                                wavelength, angle, tauC,
                                transmission_double_chopper,
                                transmission_single_chopper, penetration_depth,
                                resolution_double_chopper,
                                resolution_single_chopper, slit_optimiser,
                                beamfrac, beamfrackernel,
                                height_of_beam_after_dx,
                                energy_wavelength, xray_energy,
                                xray_wavelength, velocity_wavelength,
                                wavelength_velocity, wavelength_energy,
                                actual_footprint)
from refnx.util.quickplot import refplot

from refnx._lib._testutils import PytestTester

test = PytestTester(__name__)
del PytestTester


__all__ = [s for s in dir() if not s.startswith('_')]



import os

from refnx.reflect.reflect_model import (ReflectModel, reflectivity,
                                         MixedReflectModel)
from refnx.reflect.structure import (Structure, SLD, Slab, Component,
                                     sld_profile, Stack)
from refnx.reflect.interface import (Erf, Interface, Linear, Exponential,
                                     Tanh, Sinusoidal, Step)
from refnx.reflect.spline import Spline
from refnx.reflect._lipid import LipidLeaflet
from refnx._lib._testutils import PytestTester
from refnx.reflect._app import gui, main


# On OSX, we can get a runtime error due to multiple OpenMP libraries loaded
# simultaneously. This can happen for instance when calling BLAS inside a
# prange. Setting the following environment variable allows multiple OpenMP
# libraries to be loaded.
# OMP
os.environ.setdefault('KMP_DUPLICATE_LIB_OK', 'True')

try:
    from refnx.reflect._interactive_modeller import Motofit
except ImportError:

    class Motofit():
        def __init__(self):
            raise RuntimeError("To run Motofit you need to install"
                               " IPython, ipywidgets, traitlets,"
                               " matplotlib")

        def __call__(self, dummy):
            pass


test = PytestTester(__name__)
del PytestTester


__all__ = [s for s in dir() if not s.startswith('_')]



import os.path
import os
import logging
from copy import copy

from PyQt5 import QtCore
import numpy as np

from refnx._lib import preserve_cwd
from refnx.dataset import ReflectDataset
from refnx.reduce import (number_datafile, PlatypusReduce, basename_datafile,
                          PlatypusNexus)


reducer_entry = [('use', False), ('scale', 1.), ('reflect-1', ''),
                 ('reflect-2', ''), ('reflect-3', ''), ('direct-1', ''),
                 ('direct-2', ''), ('direct-3', ''), ('flood', '')]
default_reducer_entry = {re[0]: re[1] for re in reducer_entry}


class ReductionState(object):
    """
    Reduces multiple reflectometry datafiles. Attributes control how the
    reduction proceeds.
    """
    def __init__(self, manual_beam_finder=None):
        super(ReductionState, self).__init__()

        # Each entry specifies the scale factor, whether the file is used for
        # reduction and the reflected and direct beam runs
        self.reduction_entries = {}

        self.default_reduction_options = {
            'low_wavelength': (2.5, float),
            'high_wavelength': (19, float),
            'rebin_percent': (2., float),
            'expected_centre': (500., float),
            'manual_beam_find': (False, bool),
            'background_subtraction': (True, bool),
            'monitor_normalisation': (True, bool),
            'save_offspecular': (True, bool),
            'save_spectrum': (True, bool),
            'streamed_reduction': (False, bool)}

        # the time slices for streamed reduction
        self.stream_start = 0
        self.stream_end = 3600
        self.stream_duration = 30

        # where the data, streamed data and output files are located
        self.data_directory = ''
        self.streamed_directory = ''
        self.output_directory = ''

        self.manual_beam_finder = manual_beam_finder

        for attr, val in self.default_reduction_options.items():
            setattr(self, attr, val[0])

        self.peak_pos = (90, 5)

        self.save_state_path = None

    def __getstate__(self):
        d = self.__dict__
        e = copy(d)
        e.pop('manual_beam_finder')
        return e

    @preserve_cwd
    def reducer(self, callback=None):
        """
        Reduce all the entries in reduction_entries

        Parameters
        ----------
        callback : callable
            Function, `f(percent_finished)` that is called with the current
            percentage progress of the reduction
        """

        # refnx.reduce.reduce needs you to be in the directory where you're
        # going to write files to
        if self.output_directory:
            os.chdir(self.output_directory)

        # if no data directory was specified then assume it's the cwd
        data_directory = self.data_directory
        if not data_directory:
            data_directory = './'

        def full_path(fname):
            f = os.path.join(data_directory, fname)
            return f

        # if the streamed directory isn't mentioned then assume it's the same
        # as the data directory
        streamed_directory = self.streamed_directory
        if not os.path.isdir(streamed_directory):
            self.streamed_directory = data_directory

        logging.info('-------------------------------------------------------'
                     '\nStarting reduction run')
        logging.info(
            'data_folder={data_directory}, trim_trailing=True, '
            'lo_wavelength={low_wavelength}, '
            'hi_wavelength={high_wavelength}, '
            'rebin_percent={rebin_percent}, '
            'normalise={monitor_normalisation}, '
            'background={background_subtraction} '
            'eventmode={streamed_reduction} '
            'event_folder={streamed_directory}'.format(**self.__dict__))

        # sets up time slices for event reduction
        if self.streamed_reduction:
            eventmode = np.arange(self.stream_start,
                                  self.stream_end,
                                  self.stream_duration)
            eventmode = np.r_[eventmode, self.stream_end]
        else:
            eventmode = None

        # are you manual beamfinding?
        peak_pos = None
        if (self.manual_beam_find and
                self.manual_beam_finder is not None):
            peak_pos = -1

        idx = 0

        cached_direct_beams = {}

        for row, val in self.reduction_entries.items():
            if not val['use']:
                continue

            flood = None
            if val['flood']:
                flood = full_path(val['flood'])

            combined_dataset = None

            # process entries one by one
            for ref, db in zip(['reflect-1', 'reflect-2', 'reflect-3'],
                               ['direct-1', 'direct-2', 'direct-3']):
                reflect = val[ref]
                direct = val[db]

                # if the file doesn't exist there's no point continuing
                if ((not os.path.isfile(full_path(reflect))) or
                        (not os.path.isfile(full_path(direct)))):
                    continue

                # which of the nspectra to reduce (or all)
                ref_pn = PlatypusNexus(full_path(reflect))

                if direct not in cached_direct_beams:
                    cached_direct_beams[direct] = PlatypusReduce(
                        direct,
                        data_folder=data_directory)

                reducer = cached_direct_beams[direct]

                try:
                    reduced = reducer(
                        ref_pn, scale=val['scale'],
                        h5norm=flood,
                        lo_wavelength=self.low_wavelength,
                        hi_wavelength=self.high_wavelength,
                        rebin_percent=self.rebin_percent,
                        normalise=self.monitor_normalisation,
                        background=self.background_subtraction,
                        manual_beam_find=self.manual_beam_finder,
                        peak_pos=peak_pos,
                        eventmode=eventmode,
                        event_folder=streamed_directory)
                except Exception as e:
                    # typical Exception would be ValueError for non overlapping
                    # angles
                    logging.info(e)
                    continue

                logging.info(
                    'Reduced {} vs {}, scale={}, angle={}'.format(
                        reflect, direct, val['scale'],
                        reduced[1]['omega'][0, 0]))

                if combined_dataset is None:
                    combined_dataset = ReflectDataset()

                    fname = basename_datafile(reflect)
                    fname_dat = os.path.join(self.output_directory,
                                             'c_{0}.dat'.format(fname))
                    fname_xml = os.path.join(self.output_directory,
                                             'c_{0}.xml'.format(fname))

                try:
                    combined_dataset.add_data(reducer.data(),
                                              requires_splice=True,
                                              trim_trailing=True)
                except ValueError as e:
                    # datasets don't overlap
                    logging.info(e)
                    continue

            if combined_dataset is not None:
                # after you've finished reducing write a combined file.
                with open(fname_dat, 'wb') as f:
                    combined_dataset.save(f)
                with open(fname_xml, 'wb') as f:
                    combined_dataset.save_xml(f)
                logging.info(
                    'Written combined files: {} and {}'.format(
                        fname_dat, fname_xml))

            # can be used to create a progress bar
            idx += 1
            if callback is not None:
                ok = callback(100 * idx / len(self.reduction_entries))
                if not ok:
                    break

        logging.info('\nFinished reduction run'
                     '-------------------------------------------------------')


class ReductionTableModel(QtCore.QAbstractTableModel):
    '''
        a model for displaying in a QtGui.QTableView
    '''
    def __init__(self, reduction_state, parent=None):
        super(ReductionTableModel, self).__init__(parent)
        self._reduction_state = reduction_state

    @property
    def reduction_state(self):
        return self._reduction_state()

    def rowCount(self, parent=QtCore.QModelIndex()):
        return 200

    def columnCount(self, parent=QtCore.QModelIndex()):
        return len(reducer_entry)

    def headerData(self, section, orientation, role=QtCore.Qt.DisplayRole):
        """ Set the headers to be displayed. """
        if role != QtCore.Qt.DisplayRole:
            return None

        if orientation == QtCore.Qt.Vertical:
            return None

        if orientation == QtCore.Qt.Horizontal:
            return reducer_entry[section][0]

        return None

    def flags(self, index):
        # row = index.row()
        col = index.column()
        if not col:
            return (QtCore.Qt.ItemIsUserCheckable |
                    QtCore.Qt.ItemIsEnabled)

        return (QtCore.Qt.ItemIsEditable |
                QtCore.Qt.ItemIsEnabled |
                QtCore.Qt.ItemIsSelectable)

    def data(self, index, role=QtCore.Qt.DisplayRole):
        if not index.isValid():
            return False

        row = index.row()
        col = index.column()

        state = self.reduction_state
        attr_name = reducer_entry[col][0]

        if row in state.reduction_entries:
            entry = state.reduction_entries[row]
        else:
            entry = default_reducer_entry

        value = entry[attr_name]

        if role == QtCore.Qt.CheckStateRole:
            if not col and value:
                return QtCore.Qt.Checked
            elif not col:
                return QtCore.Qt.Unchecked

        if role == QtCore.Qt.DisplayRole and col:
            if attr_name == 'scale':
                return str(float(value))
            else:
                return value

    def setData(self, index, value, role=QtCore.Qt.EditRole):
        row = index.row()
        col = index.column()

        if not index.isValid():
            return False

        state = self.reduction_state
        attr_name = reducer_entry[col][0]

        if row in state.reduction_entries:
            entry = state.reduction_entries[row]
        else:
            entry = {re[0]: re[1] for re in reducer_entry}
            state.reduction_entries[row] = entry

        if role == QtCore.Qt.CheckStateRole and col == 0:
            entry['use'] = (value == QtCore.Qt.Checked)

        if role == QtCore.Qt.EditRole:
            if col == 0:
                save_value = False
                if value == QtCore.Qt.Checked:
                    save_value = True
            elif col == 1:
                try:
                    save_value = float(value)
                except ValueError:
                    save_value = 1
            else:
                # you're editing reflect/direct beam names
                try:
                    save_value = int(value)
                    save_value = number_datafile(save_value)
                except ValueError:
                    if value:
                        if value.endswith('.nx.hdf'):
                            save_value = value
                        else:
                            save_value = value + '.nx.hdf'
                    else:
                        return False

            entry[attr_name] = save_value

        self.dataChanged.emit(index, index)
        return True



"""
Interfacial models

Definitions taken from:

Svechnikov, M.; Pariev, D.; Nechay, A.; Salashchenko, N.; Chkhalo, N.;
Vainer, Y. & Gaman, D., "Extended model for the reconstruction of periodic
multilayers from extreme ultraviolet and X-ray reflectivity data",
Journal of Applied Crystallography, 2017, 50, 1428-1440

The Sinusoidal and Exponential definitions are incorrect in that paper
though. The correct equations are in:
Stearns, D. G. J. Appl. Phys., 1989, 65, 491–506.

The tanh definition has various definitions. This is taken from:
D. Bahr; W. Press; R. Jebasinski; S. Mantl, Phys. Rev. B, 1993, 47 (8), 4385
"""
import numpy as np
from scipy.stats import norm


_SQRT3 = np.sqrt(3.)
_SQRT2 = np.sqrt(2.)
_GAMMA = np.pi / np.sqrt(np.pi * np.pi - 8.)


class Interface(object):
    """
    Defines an Interfacial profile
    """
    def __init__(self):
        pass

    def __call__(self, z, scale=1, loc=0):
        raise NotImplementedError("You can't use the Interface superclass to"
                                  "calculate profiles")


class Erf(Interface):
    """
    An Error function interfacial profile

    Notes
    -----
    Svechnikov, M.; Pariev, D.; Nechay, A.; Salashchenko, N.; Chkhalo, N.;
    Vainer, Y. & Gaman, D., "Extended model for the reconstruction of periodic
    multilayers from extreme ultraviolet and X-ray reflectivity data",
    Journal of Applied Crystallography, 2017, 50, 1428-1440
    """
    def __init__(self):
        super(Erf, self).__init__()

    def __call__(self, z, scale=1, loc=0):
        return norm.cdf(z, scale=scale, loc=loc)

    def __repr__(self):
        return "Erf()"


class Linear(Interface):
    """
    A Linear function interfacial profile

    Notes
    -----
    Stearns, D. G. J. Appl. Phys., 1989, 65, 491–506.
    """
    def __init__(self):
        super(Linear, self).__init__()

    def __call__(self, z, scale=1, loc=0):
        new_z = z - loc
        f = 0.5 + new_z / (2 * _SQRT3 * scale)
        f[new_z <= -_SQRT3 * scale] = 0
        f[new_z >= _SQRT3 * scale] = 1
        return f

    def __repr__(self):
        return "Linear()"


class Exponential(Interface):
    """
    An Exponential interfacial profile

    Notes
    -----
    Stearns, D. G. J. Appl. Phys., 1989, 65, 491–506.
    """
    def __init__(self):
        super(Exponential, self).__init__()

    def __call__(self, z, scale=1, loc=0):
        new_z = z - loc
        f = np.zeros_like(new_z)
        f[new_z > 0] = 1 - 0.5 * np.exp(-_SQRT2 * new_z[new_z > 0] / scale)
        f[new_z <= 0] = 0.5 * np.exp(_SQRT2 * new_z[new_z <= 0] / scale)
        return f

    def __repr__(self):
        return "Exponential()"


class Tanh(Interface):
    """
    A hyperbolic tangent (tanh) interfacial profile

    Notes
    -----
    D. Bahr; W. Press; R. Jebasinski; S. Mantl,
    Phys. Rev. B,1993, 47 (8), 4385
    """
    def __init__(self):
        super(Tanh, self).__init__()

    def __call__(self, z, scale=1, loc=0):
        arg = np.sqrt(2 / np.pi) * (z - loc) / scale
        return 0.5 * (1 + np.tanh(arg))

    def __repr__(self):
        return "Tanh()"


class Sinusoidal(Interface):
    """
    A sinusoidal (sin) interfacial profile

    Notes
    -----
    Stearns, D. G. J. Appl. Phys., 1989, 65, 491–506.
    """
    def __init__(self):
        super(Sinusoidal, self).__init__()

    def __call__(self, z, scale=1, loc=0):
        new_z = z - loc
        f = 0.5 + 0.5 * np.sin(np.pi * new_z / _GAMMA / 2. / scale)
        f[new_z <= -_GAMMA * scale] = 0
        f[new_z >= _GAMMA * scale] = 1
        return f

    def __repr__(self):
        return "Sinusoidal()"


class Step(Interface):
    """
    A step function interfacial profile

    Notes
    -----
    Svechnikov, M.; Pariev, D.; Nechay, A.; Salashchenko, N.; Chkhalo, N.;
    Vainer, Y. & Gaman, D., "Extended model for the reconstruction of periodic
    multilayers from extreme ultraviolet and X-ray reflectivity data",
    Journal of Applied Crystallography, 2017, 50, 1428-1440
    """
    def __init__(self):
        super(Step, self).__init__()

    def __call__(self, z, scale=1, loc=0):
        new_z = z - loc
        f = np.ones_like(new_z) * 0.5
        f[new_z <= -scale] = 0
        f[new_z >= scale] = 1
        return f

    def __repr__(self):
        return "Step()"



import pickle

import numpy as np
from numpy.testing import (assert_almost_equal, assert_equal, assert_,
                           assert_allclose)

from refnx.analysis import Parameter, Model


def line(x, params, *args, **kwds):
    p_arr = np.array(params)
    return p_arr[0] + x * p_arr[1]


def line2(x, p):
    return p['c'].value + p['m'].value * x


def line3(x, params, x_err=None):
    pass


class TestModel(object):

    def setup_method(self):
        pass

    def test_evaluation(self):
        c = Parameter(1.0, name='c')
        m = Parameter(2.0, name='m')
        p = c | m

        fit_model = Model(p, fitfunc=line)
        x = np.linspace(0, 100., 20)
        y = 2. * x + 1.

        # different ways of getting the model instance to evaluate
        assert_equal(fit_model.model(x, p), y)
        assert_equal(fit_model(x, p), y)
        assert_equal(fit_model.model(x), y)
        assert_equal(fit_model(x), y)

        # can we pickle the model object
        pkl = pickle.dumps(fit_model)
        unpkl = pickle.loads(pkl)
        assert_equal(unpkl(x), y)

        # you should be able to use a lambda
        fit_model = Model(p, fitfunc=line2)
        assert_equal(fit_model(x, p), y)

        # and swap the order of parameters - retrieve by key
        p = m | c
        fit_model = Model(p, fitfunc=line2)
        assert_equal(fit_model(x, p), y)

    def test_xerr(self):
        c = Parameter(1.0, name='c')
        m = Parameter(2.0, name='m')
        p = c | m

        fit_model = Model(p, fitfunc=line3)
        assert_(fit_model._fitfunc_has_xerr is True)

        fit_model = Model(p, fitfunc=line2)
        assert_(fit_model._fitfunc_has_xerr is False)



import os.path
from copy import deepcopy
import pickle
import os
import sys
import time
import csv
from multiprocessing import get_context

import numpy as np
import matplotlib

from PyQt5 import QtCore, QtGui, QtWidgets, uic

from matplotlib.backends.backend_qt5agg import (
    FigureCanvasQTAgg as FigureCanvas)
from matplotlib.backends.backend_qt5agg import (
    NavigationToolbar2QT as NavigationToolbar)
from matplotlib.figure import Figure
import matplotlib.artist as artist
import matplotlib.lines as lines


from .SLD_calculator_view import SLDcalculatorView
from .datastore import DataStore
from .treeview_gui_model import (TreeModel, Node, DatasetNode, DataObjectNode,
                                 ComponentNode, StructureNode, PropertyNode,
                                 ReflectModelNode, ParNode, TreeFilter,
                                 find_data_object, SlabNode, StackNode)
from ._lipid_leaflet import LipidLeafletDialog
from ._optimisation_parameters import OptimisationParameterView
from ._spline import SplineDialog
from ._mcmc import (ProcessMCMCDialog, SampleMCMCDialog, _plots)

import refnx
from refnx.analysis import (CurveFitter, Objective,
                            Transform, GlobalObjective)
from refnx.reflect import SLD, ReflectModel, Slab, Stack, Structure
from refnx.dataset import Data1D
from refnx.reflect._code_fragment import code_fragment
from refnx._lib import unique, flatten, MapWrapper


# matplotlib.use('Qt5Agg')
UI_LOCATION = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                           'ui')


class MotofitMainWindow(QtWidgets.QMainWindow):
    """
    Main View window for Motofit
    """

    def __init__(self, parent=None):
        super(MotofitMainWindow, self).__init__(parent)

        # load the GUI from the ui file
        self.ui = uic.loadUi(os.path.join(UI_LOCATION, 'motofit.ui'),
                             self)

        self.error_handler = QtWidgets.QErrorMessage()

        #######################################################################
        # Everything ending in 'model' refers to a QtAbstract<x>Model.  These
        # are the basis for GUI elements.  They also contain data.
        data_container = DataStore()

        # a flag that is used by update_gui_model to decide whether to redraw
        # object graphs in response to the treeModel being changed
        self._hold_updating = False

        # set up tree view
        self.treeModel = TreeModel(data_container)

        # the filter controls what rows are presented in the treeView
        self.treeFilter = TreeFilter(self.treeModel)
        self.treeFilter.setSourceModel(self.treeModel)
        self.ui.treeView.setModel(self.treeFilter)

        self.ui.treeView.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)

        # context menu for the treeView
        self.context_menu = OpenMenu(self.ui.treeView)

        self.ui.treeView.customContextMenuRequested.connect(self.context_menu)
        self.context_menu.add_to_fit_action.triggered.connect(
            self.on_add_to_fit_button_clicked)
        self.context_menu.remove_from_fit_action.triggered.connect(
            self.on_remove_from_fit_action)
        self.context_menu.link_action.triggered.connect(
            self.link_action)
        self.context_menu.link_equivalent_action.triggered.connect(
            self.link_equivalent_action)
        self.context_menu.unlink_action.triggered.connect(
            self.unlink_action)
        self.context_menu.copy_from_action.triggered.connect(
            self.copy_from_action)
        self.context_menu.add_mixed_area.triggered.connect(
            self.add_mixed_area_action)
        self.context_menu.remove_mixed_area.triggered.connect(
            self.remove_mixed_area_action)
        self.actionLink_Selected.triggered.connect(self.link_action)
        self.actionUnlink_selected_Parameters.triggered.connect(
            self.unlink_action)
        self.actionLink_Equivalent_Parameters.triggered.connect(
            self.link_equivalent_action)

        self.treeModel.dataChanged.connect(self.tree_model_data_changed)
        self.treeModel.rowsRemoved.connect(self.tree_model_structure_changed)
        self.treeModel.rowsMoved.connect(self.tree_model_structure_changed)
        self.treeModel.rowsInserted.connect(self.tree_model_structure_changed)

        # list view for datasets being fitted
        self.currently_fitting_model = CurrentlyFitting(self)
        self.ui.currently_fitting.setModel(self.currently_fitting_model)
        #######################################################################

        # attach the reflectivity graphs and the SLD profiles
        self.modify_gui()

        # holds miscellaneous information on program settings
        self.settings = ProgramSettings()
        self.settings.current_dataset_name = 'theoretical'

        theoretical = data_container['theoretical']
        model = theoretical.model
        dataset = theoretical.dataset
        resolution = self.settings.resolution / 100.
        transform = Transform(self.settings.transformdata)
        fit = model(dataset.x, x_err=dataset.x * resolution)
        fit, _ = transform(dataset.x, fit)
        sld = model.structure.sld_profile()

        graph_properties = theoretical.graph_properties
        graph_properties['ax_fit'] = self.reflectivitygraphs.axes[0].plot(
            dataset.x,
            fit, color='r',
            linestyle='-', lw=1,
            label='theoretical',
            picker=5)[0]

        graph_properties['ax_sld_profile'] = self.sldgraphs.axes[0].plot(
            sld[0],
            sld[1],
            linestyle='-', color='r')[0]

        self.restore_settings()

        self.spline_dialog = SplineDialog(self)
        self.sld_calculator = SLDcalculatorView(self)
        self.lipid_leaflet = LipidLeafletDialog(self)
        self.optimisation_parameters = OptimisationParameterView(self)
        self.data_object_selector = DataObjectSelectorDialog(self)

        self.ui.treeView.setColumnWidth(0, 200)
        h = self.ui.treeView.header()
        h.setMinimumSectionSize(100)
        # h.setSectionResizeMode(0, QtWidgets.QHeaderView.ResizeToContents)

        # redirect stdout to a console window
        console = EmittingStream()
        sys.stdout = console
        console.textWritten.connect(self.writeTextToConsole)

        print('Session started at:', time.asctime(time.localtime(time.time())))

    def __del__(self):
        # Restore sys.stdout
        sys.stdout = sys.__stdout__

    @QtCore.pyqtSlot(QtGui.QDropEvent)
    def dropEvent(self, event):
        m = event.mimeData()
        urls = m.urls()

        # convert url's to files
        urls_as_files = [url.toLocalFile() for url in urls]

        # if you've only got one url then try and load it as an experiment
        if len(urls_as_files) == 1:
            try:
                self._restore_state(urls_as_files[0])
                return
            except Exception:
                pass

        # then try and load urls as data files
        try:
            loaded_names = self.load_data(urls_as_files)
        except Exception:
            loaded_names = []

        # then try and load the remainder as models
        remainder_urls = set(urls_as_files).difference(set(loaded_names))
        for url in remainder_urls:
            try:
                self.load_model(url)
                continue
            except Exception:
                pass

    @QtCore.pyqtSlot(QtGui.QDragEnterEvent)
    def dragEnterEvent(self, event):
        m = event.mimeData()
        if m.hasUrls():
            event.acceptProposedAction()

    def writeTextToConsole(self, text):
        self.ui.console_text_edit.moveCursor(QtGui.QTextCursor.End)
        self.ui.console_text_edit.insertPlainText(text)

    def _saveState(self, experiment_file_name):
        state = {}
        self.settings.experiment_file_name = experiment_file_name
        state['datastore'] = self.treeModel.datastore
        state['history'] = self.ui.console_text_edit.toPlainText()
        state['settings'] = self.settings
        state['refnx.version'] = refnx.version.version

        fit_list = self.currently_fitting_model
        state['currently_fitting'] = fit_list.datasets

        with open(os.path.join(experiment_file_name), 'wb') as f:
            pickle.dump(state, f, -1)

        self.setWindowTitle('Motofit - ' + experiment_file_name)

    @QtCore.pyqtSlot()
    def on_actionSave_File_triggered(self):
        if os.path.isfile(self.settings.experiment_file_name):
            self._saveState(self.settings.experiment_file_name)
        else:
            self.on_actionSave_File_As_triggered()

    @QtCore.pyqtSlot()
    def on_actionSave_File_As_triggered(self):
        experiment_file_name, ok = QtWidgets.QFileDialog.getSaveFileName(
            self, 'Save experiment as:', 'experiment.mtft')

        if not ok:
            return

        path, ext = os.path.splitext(experiment_file_name)
        if ext != '.mtft':
            experiment_file_name = path + '.mtft'

        self._saveState(experiment_file_name)

    def _restore_state(self, experiment_file_name):
        with open(experiment_file_name, 'rb') as f:
            state = pickle.load(f)

        if not state:
            print("Couldn't load experiment")
            return

        try:
            self.treeModel._data = state[
                'datastore']
            self.treeModel.rebuild()

            # amend the internal state to compensate for mtft files saved in
            # older versions missing attributes saved in later versions.
            self.compensate_older_versions()

            # remove and add datasetsToGraphs
            self.reflectivitygraphs.remove_traces()
            self.sldgraphs.remove_traces()
            ds = [d for d in self.treeModel.datastore]
            self.add_data_objects_to_graphs(ds)
            self.update_gui_model(ds)
            self.reflectivitygraphs.draw()
        except Exception as e:
            version = state.get('refnx.version', 'N/A')
            msg("Failed to load experiment. It may have been saved in a"
                " previous refnx version ({}). Please use that version to"
                " continue with analysis, refnx will now"
                " close.".format(version))
            raise e

        try:
            self.ui.console_text_edit.setPlainText(state['history'])
            self.settings = state['settings']
            self.settings.experiment_file_name = experiment_file_name
            self.restore_settings()

            while self.data_object_selector.data_objects.count():
                self.data_object_selector.data_objects.takeItem(0)

            self.data_object_selector.addItems(self.treeModel.datastore.names)

            self.currently_fitting_model = CurrentlyFitting(self)
            fit_list = self.currently_fitting_model
            self.ui.currently_fitting.setModel(fit_list)
            fit_list.addItems(state['currently_fitting'])

        except KeyError as e:
            print(repr(e))
            return

    def restore_settings(self):
        """
        applies the program settings to the GUI
        """
        title = 'Motofit'
        if len(self.settings.experiment_file_name):
            title += ' - ' + self.settings.experiment_file_name
        self.setWindowTitle(title)

        self.select_fitting_algorithm(self.settings.fitting_algorithm)
        self.ui.use_errors_checkbox.setChecked(self.settings.useerrors)
        self.settransformoption(self.settings.transformdata)

    def compensate_older_versions(self):
        """
        Amends the internal state of the program to add attributes that may
        be missing in experiment files that were saved in earlier versions of
        the GUI.
        """
        # add interfaces attribute to all Components (added in v0.1.8)
        # another way of doing it would be to use
        # `self.__dict__.get('_interfaces', None)` in the interfaces property,
        # but that is slower.
        for do in self.treeModel.datastore:
            model = do.model
            s = model.structure
            for component in s:
                if not hasattr(component, '_interfaces'):
                    component._interfaces = None

    def apply_settings_to_params(self, params):
        for key in self.settings.__dict__:
            params[key] = self.settings[key]

    @QtCore.pyqtSlot()
    def on_actionLoad_File_triggered(self):
        experimentFileName, ok = QtWidgets.QFileDialog.getOpenFileName(
            self,
            caption='Select Experiment File',
            filter='Experiment Files (*.mtft)')
        if not ok:
            return

        self._restore_state(experimentFileName)

    def load_data(self, files):
        """
        Loads datasets into the app

        Parameters
        ----------
        files : sequence
            Sequence of `str` specifying filenames to load the datasets from

        Returns
        -------
        fnames : sequence
            Sequence of `str` containing the filenames of the successfully
            loaded datasets.
        """
        datastore = self.treeModel.datastore

        existing_data_objects = datastore.names

        data_objects = []
        fnames = []
        for file in files:
            if os.path.isfile(file):
                try:
                    data_object = self.treeModel.load_data(file)
                    if data_object is not None:
                        data_objects.append(data_object)
                        fnames.append(file)
                except Exception:
                    continue

        loaded_data_objects = [data_object.name for data_object
                               in data_objects]
        new_names = [n for n in loaded_data_objects if
                     n not in existing_data_objects]

        # give a newly loaded data object a simple model. You don't want to do
        # this for an object that's already been loaded.
        for name in new_names:
            fronting = SLD(0, name='fronting')
            sio2 = SLD(3.47, name='1')
            backing = SLD(2.07, name='backing')
            s = fronting() | sio2(15, 3) | backing(0, 3)
            data_object_node = self.treeModel.data_object_node(name)
            model = ReflectModel(s)
            model.name = name
            data_object_node.set_reflect_model(model)

        # for the intersection of loaded and old, refresh the plot.
        refresh_names = [n for n in existing_data_objects if
                         n in loaded_data_objects]

        refresh_data_objects = [datastore[name] for name
                                in refresh_names]
        self.redraw_data_object_graphs(refresh_data_objects)

        # for totally new, then add to graphs
        new_data_objects = [datastore[name] for name in new_names]
        self.add_data_objects_to_graphs(new_data_objects)

        self.calculate_chi2(data_objects)

        # add newly loads to the data object selector dialogue
        self.data_object_selector.addItems(new_names)
        return fnames

    @QtCore.pyqtSlot()
    def on_actionLoad_Data_triggered(self):
        """
        you load data
        """
        files = QtWidgets.QFileDialog.getOpenFileNames(
            self,
            caption='Select Reflectivity Files')

        if files:
            self.load_data(files[0])

    @QtCore.pyqtSlot()
    def on_actionRemove_Data_triggered(self):
        """
        you remove data
        """
        # retrieve data_objects that need to be removed
        datastore = self.treeModel.datastore

        self.data_object_selector.setWindowTitle('Select datasets to remove')
        ok = self.data_object_selector.exec_()
        if not ok:
            return
        items = self.data_object_selector.data_objects.selectedItems()
        names = [item.text() for item in items]
        if 'theoretical' in names:
            names.pop(names.index('theoretical'))

        fit_list = self.currently_fitting_model

        for which_dataset in names:
            # remove from list of datasets to be fitted, if present
            if which_dataset in fit_list.datasets:
                fit_list.removeItems([fit_list.datasets.index(which_dataset)])

            self.reflectivitygraphs.remove_trace(
                datastore[which_dataset])
            self.sldgraphs.remove_trace(datastore[which_dataset])
            self.treeModel.remove_data_object(which_dataset)

            # remove from data object selector
            self.data_object_selector.removeItem(which_dataset)

    @QtCore.pyqtSlot()
    def on_actionSave_Fit_triggered(self):
        datastore = self.treeModel.datastore

        self.data_object_selector.setWindowTitle('Select fits to save')
        ok = self.data_object_selector.exec_()
        if not ok:
            return
        items = self.data_object_selector.data_objects.selectedItems()
        names = [item.text() for item in items]
        dialog = QtWidgets.QFileDialog(self)
        dialog.setFileMode(QtWidgets.QFileDialog.Directory)
        if dialog.exec_():
            folder = dialog.selectedFiles()
            for name in names:
                datastore[name].save_fit(
                    os.path.join(folder[0], 'fit_' + name + '.dat'))

    @QtCore.pyqtSlot()
    def on_actionSave_SLD_Curve_triggered(self):
        # saves an SLD curve as a text file
        datastore = self.treeModel.datastore

        self.data_object_selector.setWindowTitle('Select fits to save')
        ok = self.data_object_selector.exec_()
        if not ok:
            return
        items = self.data_object_selector.data_objects.selectedItems()
        names = [item.text() for item in items]
        dialog = QtWidgets.QFileDialog(self)
        dialog.setFileMode(QtWidgets.QFileDialog.Directory)
        if dialog.exec_():
            folder = dialog.selectedFiles()
            for name in names:
                data_object = datastore[name]
                sld_curve = np.array(data_object.sld_profile).T
                if sld_curve is not None:
                    # it may be None if it's a mixed area model
                    sld_file_name = os.path.join(folder[0],
                                                 'sld_' + name + '.dat')
                    np.savetxt(sld_file_name, sld_curve)

    def load_model(self, model_file_name):
        with open(model_file_name, 'rb') as f:
            model = pickle.load(f)

        if not isinstance(model, ReflectModel):
            msg('The pkl file you were loading was not a ReflectModel')
            return

        data_object_node = self.treeModel.data_object_node(model.name)
        if data_object_node is not None:
            data_object_node.set_reflect_model(model)
        else:
            # there is not dataset with that name, put the model over the
            # theoretical one
            data_object_node = self.treeModel.data_object_node('theoretical')
            data_object_node.set_reflect_model(model)
        self.update_gui_model([data_object_node.data_object])

    @QtCore.pyqtSlot()
    def on_actionLoad_Model_triggered(self):
        # load a model from a pickle file
        model_file_name, ok = QtWidgets.QFileDialog.getOpenFileName(
            self, 'Select Model File')
        if not ok:
            return
        self.load_model(model_file_name)

    @QtCore.pyqtSlot()
    def on_actionSave_Model_triggered(self):
        # save a model to a pickle file
        # which model are you saving?
        datastore = self.treeModel.datastore

        self.data_object_selector.setWindowTitle('Select models to save')
        ok = self.data_object_selector.exec_()
        if not ok:
            return
        items = self.data_object_selector.data_objects.selectedItems()
        names = [item.text() for item in items]

        dialog = QtWidgets.QFileDialog(self)
        dialog.setFileMode(QtWidgets.QFileDialog.Directory)
        dialog.setWindowTitle('Where do you want to save the models?')
        if dialog.exec_():
            folder = dialog.selectedFiles()

            for name in names:
                fname = os.path.join(folder[0],
                                     'coef_' + name + '.pkl')
                model = datastore[name]
                model.save_model(fname)

    @QtCore.pyqtSlot()
    def on_actionProcess_MCMC_triggered(self):
        """
        Process an MCMC chain. Can only do against current fitting setup though
        """
        names_to_fit = self.currently_fitting_model.datasets
        # retrieve data_objects
        datastore = self.treeModel.datastore
        data_objects = [datastore[name] for name in names_to_fit]
        objective = self.create_objective(data_objects)

        try:
            dialog = ProcessMCMCDialog(objective, None, parent=self)
            if dialog.chain is None:
                return
            dialog.exec_()
            print(str(objective))
            _plots(objective, nplot=dialog.nplot.value(), folder=dialog.folder)
        except Exception as e:
            print(repr(e))
            msg("MCMC processing went wrong. The MCMC chain can only be"
                " processed against the fitting setup that created it.")

    @QtCore.pyqtSlot()
    def on_actionExport_parameters_triggered(self):
        # save all parameter values to a text file
        datastore = self.treeModel.datastore

        self.data_object_selector.setWindowTitle('Select parameters to export')
        ok = self.data_object_selector.exec_()
        if not ok:
            return
        items = self.data_object_selector.data_objects.selectedItems()
        names = [item.text() for item in items]

        suggested_name = os.path.join(os.getcwd(),
                                      'coefficients.csv')
        fname, ok = QtWidgets.QFileDialog.getSaveFileName(
            self,
            'Exported file name:',
            suggested_name)
        if not ok:
            return

        with open(fname, 'w', newline='') as csvfile:
            writer = csv.writer(csvfile, delimiter=',',
                                quotechar='|', quoting=csv.QUOTE_MINIMAL)
            for name in names:
                model = datastore[name].model
                writer.writerow([name])
                writer.writerow([p.name for p in flatten(model.parameters)])
                writer.writerow([p.value for p in flatten(model.parameters)])
                writer.writerow([p.stderr for p in flatten(model.parameters)])

    @QtCore.pyqtSlot()
    def on_actionExport_Code_Fragment_triggered(self):
        # exports an executable script for the current fitting system.
        # this script can be used for e.g. MCMC sampling.
        names_to_fit = self.currently_fitting_model.datasets

        if not names_to_fit:
            return

        # retrieve data_objects
        datastore = self.treeModel.datastore
        data_objects = [datastore[name] for name in names_to_fit]

        objective = self.create_objective(data_objects)
        code = code_fragment(objective)

        suggested_name = os.path.join(os.getcwd(), 'mcmc.py')
        modelFileName, ok = QtWidgets.QFileDialog.getSaveFileName(
            self,
            'Save code fragment as:',
            suggested_name)
        if not ok:
            return

        try:
            with open(modelFileName, 'w') as f:
                f.write(code)
        except Exception as e:
            print(e)

    def select_fitting_algorithm(self, method):
        meth = {'LM': self.ui.actionLevenberg_Marquardt,
                'MCMC': self.ui.actionMCMC,
                'L-BFGS-B': self.ui.actionL_BFGS_B,
                'SHGO': self.ui.actionSHGO,
                'dual_annealing': self.ui.actionDual_Annealing,
                'DE': self.ui.actionDifferential_Evolution
                }
        self.settings.fitting_algorithm = method
        meth[method].setChecked(True)
        meth.pop(method)
        for k, v in meth.items():
            v.setChecked(False)

    @QtCore.pyqtSlot()
    def on_actionDifferential_Evolution_triggered(self):
        self.select_fitting_algorithm('DE')

    @QtCore.pyqtSlot()
    def on_actionMCMC_triggered(self):
        self.select_fitting_algorithm('MCMC')

    @QtCore.pyqtSlot()
    def on_actionDual_Annealing_triggered(self):
        self.select_fitting_algorithm('dual_annealing')

    @QtCore.pyqtSlot()
    def on_actionSHGO_triggered(self):
        self.select_fitting_algorithm('SHGO')

    @QtCore.pyqtSlot()
    def on_actionLevenberg_Marquardt_triggered(self):
        self.select_fitting_algorithm('LM')

    @QtCore.pyqtSlot()
    def on_actionL_BFGS_B_triggered(self):
        self.select_fitting_algorithm('L-BFGS-B')

    def change_Q_range(self, qmin, qmax, numpnts):
        data_object_node = self.treeModel.data_object_node('theoretical')

        theoretical = data_object_node._data
        dataset = theoretical.dataset

        new_x = np.linspace(qmin, qmax, numpnts)
        new_y = np.zeros_like(new_x) * np.nan
        dataset.data = (new_x, new_y)
        data_object_node.set_dataset(dataset)
        self.update_gui_model([theoretical])

    @QtCore.pyqtSlot()
    def on_actionChange_Q_range_triggered(self):
        datastore = self.treeModel.datastore
        theoretical = datastore['theoretical']
        qmin = min(theoretical.dataset.x)
        qmax = max(theoretical.dataset.x)
        numpnts = len(theoretical.dataset)

        dvalidator = QtGui.QDoubleValidator(-2.0e-308, 2.0e308, 6)

        qrangeGUI = uic.loadUi(os.path.join(UI_LOCATION, 'qrangedialog.ui'))
        qrangeGUI.numpnts.setValue(numpnts)
        qrangeGUI.qmin.setValidator(dvalidator)
        qrangeGUI.qmax.setValidator(dvalidator)
        qrangeGUI.qmin.setText(str(qmin))
        qrangeGUI.qmax.setText(str(qmax))

        ok = qrangeGUI.exec_()
        if ok:
            self.change_Q_range(float(qrangeGUI.qmin.text()),
                                float(qrangeGUI.qmax.text()),
                                qrangeGUI.numpnts.value())

    @QtCore.pyqtSlot()
    def on_actionTake_Snapshot_triggered(self):
        snapshotname, ok = QtWidgets.QInputDialog.getText(
            self,
            'Take a snapshot',
            'snapshot name')
        if not ok:
            return

        # is the snapshot already present?
        datastore = self.treeModel.datastore
        snapshot_exists = datastore[snapshotname]
        data_object = self.treeModel.snapshot(snapshotname)

        if snapshot_exists is not None:
            self.redraw_data_object_graphs([data_object])
        else:
            self.add_data_objects_to_graphs([data_object])

    @QtCore.pyqtSlot()
    def on_actionResolution_smearing_triggered(self):
        currentVal = self.settings.quad_order
        value, ok = QtWidgets.QInputDialog.getInt(
            self,
            'Resolution Smearing',
            'Number of points for Gaussian Quadrature',
            currentVal,
            17)
        if not ok:
            return
        self.settings.quad_order = value
        self.update_gui_model([])

    @QtCore.pyqtSlot()
    def on_actionBatch_Fit_triggered(self):
        datastore = self.treeModel.datastore
        if len(datastore) < 2:
            return msg("You have no loaded datasets")

        alg = self.settings.fitting_algorithm
        if alg == 'MCMC':
            return msg("It's not possible to do MCMC in batch fitting mode")

        # need to retrieve the theoretical data_object because we're going to
        # use its model.
        theoretical = datastore['theoretical']

        self.data_object_selector.setWindowTitle(
            "Select datasets to batch fit (using the theoretical model)")
        ok = self.data_object_selector.exec_()
        if not ok:
            return
        items = self.data_object_selector.data_objects.selectedItems()
        names = [item.text() for item in items]

        # iterate and fit over all the selected datasets, but first copy the
        # model from the theoretical model because it's unlikely you're going
        # to setup all the individual models first.
        progress = QtWidgets.QProgressDialog('Batch fitting progress',
                                             'Stop',
                                             0,
                                             len(names),
                                             parent=self)
        progress.setWindowModality(QtCore.Qt.WindowModal)
        progress.setAutoClose(True)
        progress.setValue(0)
        progress.show()

        gui_update_list = []
        # turn of graph/treeview updating during fitting, it takes up a lot of
        # resources. Just do it all at once at the end.
        self._hold_updating = True
        last_time = time.time()
        completed = 0
        try:
            for name in names:
                if progress.wasCanceled():
                    raise StopIteration()

                data_object = datastore[name]
                if data_object.name == 'theoretical':
                    continue
                new_model = deepcopy(theoretical.model)
                new_model.name = name
                data_object_node = self.treeModel.data_object_node(name)
                data_object_node.set_reflect_model(new_model)
                successfully_fitted = self.fit_data_objects([data_object])
                if successfully_fitted:
                    gui_update_list.append(successfully_fitted)

                completed += 1
                # update progress bar every 3 secs
                if (time.time() - last_time) > 3.:
                    progress.setValue(completed)
                    last_time = time.time()

        except StopIteration:
            pass
        finally:
            self._hold_updating = False
            gui_update_list = list(flatten(gui_update_list))
            self.update_gui_fitted_data_objects(gui_update_list)
            progress.close()

    @QtCore.pyqtSlot()
    def on_actionRefresh_Data_triggered(self):
        """
            you are refreshing existing datasets
        """
        try:
            self.treeModel.refresh()
            self.redraw_data_object_graphs(None, all=True)
        except FileNotFoundError:
            print("FileNotFoundError: one or more datafiles is no longer in"
                  "their original location")
            msg("FileNotFoundError: one or more datafiles is no longer in"
                "their original location")

    @QtCore.pyqtSlot()
    def on_actionlogY_vs_X_triggered(self):
        self.settransformoption('logY')

    @QtCore.pyqtSlot()
    def on_actionY_vs_X_triggered(self):
        self.settransformoption('lin')

    @QtCore.pyqtSlot()
    def on_actionYX4_vs_X_triggered(self):
        self.settransformoption('YX4')

    @QtCore.pyqtSlot()
    def on_actionYX2_vs_X_triggered(self):
        self.settransformoption('YX2')

    def settransformoption(self, transform):
        self.ui.actionlogY_vs_X.setChecked(False)
        self.ui.actionY_vs_X.setChecked(False)
        self.ui.actionYX4_vs_X.setChecked(False)
        self.ui.actionYX2_vs_X.setChecked(False)
        if transform is None:
            self.ui.actionY_vs_X.setChecked(True)
            transform = 'lin'
        if transform == 'lin':
            self.ui.actionY_vs_X.setChecked(True)
        elif transform == 'logY':
            self.ui.actionlogY_vs_X.setChecked(True)
        elif transform == 'YX4':
            self.ui.actionYX4_vs_X.setChecked(True)
        elif transform == 'YX2':
            self.ui.actionYX2_vs_X.setChecked(True)
        self.settings.transformdata = transform

        self.redraw_data_object_graphs(None, all=True)

        # need to relimit graphs and display on a log scale if the transform
        # has changed
        self.reflectivitygraphs.axes[0].autoscale(axis='both',
                                                  tight=False,
                                                  enable=True)
        self.reflectivitygraphs.axes[0].relim()
        if transform in ['lin', 'YX2']:
            self.reflectivitygraphs.axes[0].set_yscale('log')
        else:
            self.reflectivitygraphs.axes[0].set_yscale('linear')
        self.reflectivitygraphs.draw()

    @QtCore.pyqtSlot()
    def on_actionAbout_triggered(self):
        aboutui = uic.loadUi(os.path.join(UI_LOCATION, 'about.ui'))

        licence_dir = os.path.join(UI_LOCATION, 'licences')
        licences = os.listdir(licence_dir)
        licences.remove('about')

        text = [refnx.version.version]
        with open(os.path.join(licence_dir, 'about'), 'r',
                  encoding='utf-8', errors='replace') as f:
            text.append(''.join(f.readlines()))

        for licence in licences:
            fname = os.path.join(licence_dir, licence)
            with open(fname, 'r', encoding='utf-8', errors='replace') as f:
                text.append(''.join(f.readlines()))

        display_text = '\n_______________________________________\n'.join(text)
        aboutui.textBrowser.setText(display_text)
        aboutui.exec_()

    @QtCore.pyqtSlot()
    def on_actiondocumentation_triggered(self):
        QtGui.QDesktopServices.openUrl(QtCore.QUrl(
            'https://refnx.readthedocs.io/en/latest/'))

    @QtCore.pyqtSlot()
    def on_actionAutoscale_graph_triggered(self):
        self.reflectivitygraphs.autoscale()

    @QtCore.pyqtSlot()
    def on_actionSLD_calculator_triggered(self):
        self.sld_calculator.show()

    @QtCore.pyqtSlot()
    def on_actionOptimisation_parameters_triggered(self):
        self.optimisation_parameters.show()

    @QtCore.pyqtSlot()
    def on_actionLipid_browser_triggered(self):
        self.lipid_leaflet.show()

    @QtCore.pyqtSlot()
    def on_add_layer_clicked(self):
        selected_indices = self.ui.treeView.selectedIndexes()

        if not selected_indices:
            return msg('Select a single row within a Structure to insert a'
                       ' new Component.')

        index = selected_indices[0]

        # from filter to model
        index = self.mapToSource(index)

        if not index.isValid():
            return

        item = index.internalPointer()
        hierarchy = item.hierarchy()
        # reverse so that we see bottom up. We want to see if something is a
        # Component first, then whether it's a StackNode.
        hierarchy.reverse()

        # the row you selected was within the component list
        _component = [i for i in hierarchy if (isinstance(i, ComponentNode) or
                                               isinstance(i, StackNode))]
        if not _component:
            return msg('Select a single location within a Structure to insert'
                       ' a new Component.')

        # work out which component you have.
        component = _component[0]
        host = component.parent()
        idx = component.row()
        if isinstance(host, StructureNode) and idx == len(host._data) - 1:
            return msg("You can't append a layer after the backing medium,"
                       " select a previous layer")

        # what type of component shall we add?
        comp_type = ['Slab', 'LipidLeaflet', 'Spline', 'Stack']
        which_type, ok = QtWidgets.QInputDialog.getItem(
            self, "What Component type did you want to add?", "", comp_type,
            editable=False)
        if not ok:
            return

        if which_type == 'Slab':
            c = _default_slab(parent=self)
        elif which_type == 'LipidLeaflet':
            self.lipid_leaflet.hide()
            ok = self.lipid_leaflet.exec_()
            if not ok:
                return
            c = self.lipid_leaflet.component()
        elif which_type == 'Spline':
            # if isinstance(host, StackNode):
            #     msg("Can't add Splines to a Stack")
            #     return

            ok = self.spline_dialog.exec_()
            if not ok:
                return
            c = self.spline_dialog.component()
        elif which_type == 'Stack':
            s = _default_slab()
            c = Stack(components=[s], name='Stack')

        host.insert_component(idx + 1, c)

    @QtCore.pyqtSlot()
    def on_remove_layer_clicked(self):
        selected_indices = self.ui.treeView.selectedIndexes()

        if not selected_indices:
            return msg('Select a single row within a Structure to remove.')

        index = selected_indices[0]
        if not index.isValid():
            return

        # from filter to model
        index = self.mapToSource(index)

        item = index.internalPointer()
        hierarchy = item.hierarchy()
        # reverse so that we see bottom up. We want to see if something is a
        # Component first, then whether it's a StackNode.
        hierarchy.reverse()

        # the row you selected was within the component list
        _component = [i for i in hierarchy if (isinstance(i, ComponentNode) or
                                               isinstance(i, StackNode))]
        if not _component:
            return msg('Select a single Component within a Structure to'
                       ' remove')

        # work out which component you have.
        component = _component[0]
        host = component.parent()
        idx = component.row()
        if isinstance(host, StructureNode) and idx in [0, len(host._data) - 1]:
            return msg("You can't remove the fronting or backing media")

        # TODO: unlink any parameters that might depend on this component?

        # all checking done, remove a layer
        host.remove_component(idx)

    @QtCore.pyqtSlot()
    def on_auto_limits_button_clicked(self):
        names_to_fit = self.currently_fitting_model.datasets

        datastore = self.treeModel.datastore
        data_objects = [datastore[name] for name in names_to_fit]
        # auto set the limits for the theoretical model, because it's used as
        # a springboard for the batch fit.
        data_objects.append(datastore['theoretical'])

        for data_object in data_objects:
            # retrive data_object node from the treeModel
            node = self.treeModel.data_object_node(data_object.name)
            descendants = node.descendants()
            # filter out all the parameter nodes
            par_nodes = [n for n in descendants if isinstance(n, ParNode)]
            var_par_nodes = [n for n in par_nodes if n.parameter.vary]

            for par in var_par_nodes:
                p = par.parameter
                val = p.value
                bounds = p.bounds
                if val < 0:
                    bounds.ub = 0
                    bounds.lb = 2 * val
                else:
                    bounds.lb = 0
                    bounds.ub = 2 * val

                parent, row = par.parent(), par.row()
                idx1 = self.treeModel.index(row, 3, parent.index)
                idx2 = self.treeModel.index(row, 4, parent.index)
                self.treeModel.dataChanged.emit(idx1, idx2)

    @QtCore.pyqtSlot()
    def on_add_to_fit_button_clicked(self):
        selected_indices = self.ui.treeView.selectedIndexes()

        fit_list = self.currently_fitting_model

        to_be_added = []
        for index in selected_indices:
            # from filter to model
            index = self.mapToSource(index)

            data_object_node = find_data_object(index)
            if not data_object_node:
                continue
            # see if it's already in the list
            name = data_object_node.data_object.name
            to_be_added.append(name)

        fit_list.addItems(unique(to_be_added))

    @QtCore.pyqtSlot()
    def on_remove_from_fit_button_clicked(self):
        # work out what datasets are selected in the listwidget
        # remove all those that are selected
        fit_list = self.currently_fitting_model
        selected_items = self.ui.currently_fitting.selectedIndexes()

        rows = [i.row() for i in selected_items]
        fit_list.removeItems(rows)

    def on_remove_from_fit_action(self):
        selected_indices = self.ui.treeView.selectedIndexes()
        fit_list = self.currently_fitting_model

        to_remove = []
        for index in selected_indices:
            # from filter to model
            index = self.mapToSource(index)

            data_object_node = find_data_object(index)
            if not data_object_node:
                continue
            # see if it's already in the list
            name = data_object_node.data_object.name
            if name in fit_list.datasets:
                to_remove.append(fit_list.datasets.index(name))

        fit_list.removeItems(list(unique(to_remove)))

    @QtCore.pyqtSlot()
    def on_do_fit_button_clicked(self):
        """
        you should do a fit
        """
        names_to_fit = self.currently_fitting_model.datasets

        if not names_to_fit:
            return msg("Please add datasets to fit.")

        # retrieve data_objects
        datastore = self.treeModel.datastore
        data_objects = [datastore[name] for name in names_to_fit]

        successfully_fitted = self.fit_data_objects(data_objects)
        if successfully_fitted:
            self.update_gui_fitted_data_objects(successfully_fitted)

    def create_objective(self, data_objects):
        """
        Creates an Objective for a list of DataObject.
        """

        # performs a global fit to the list of data_objects
        t = Transform(self.settings.transformdata)
        useerrors = self.settings.useerrors

        # make objectives
        objectives = []
        for data_object in data_objects:
            # this is if the user requested constant dq/q. If that's the case
            # then make a new dataset to hide the resolution information
            dataset_t = data_object.dataset
            if data_object.constantdq_q and dataset_t.x_err is not None:
                dataset_t = Data1D(data=dataset_t)
                dataset_t.x_err = None

            # can't have negative points if fitting as logR vs Q
            if self.settings.transformdata == 'logY':
                dataset_t = self.filter_neg_reflectivity(dataset_t)

            objective = Objective(data_object.model,
                                  dataset_t,
                                  name=data_object.name,
                                  transform=t,
                                  use_weights=useerrors)
            objectives.append(objective)

        if len(objectives) == 1:
            objective = objectives[0]
        else:
            objective = GlobalObjective(objectives)

        return objective

    def fit_data_objects(self, data_objects):
        """
        Simultaneously fits a sequence of datasets

        Parameters
        ----------
        data_objects: list of DataObjects

        Returns
        -------
        fitted_dataobjects : list of DataObjects
            A list of successfully fitted datasets
        """

        objective = self.create_objective(data_objects)

        alg = self.settings.fitting_algorithm

        # figure out how many varying parameters
        vp = objective.varying_parameters()
        if not vp:
            return msg("No parameters are being varied.")

        methods = {'DE': 'differential_evolution',
                   'LM': 'least_squares',
                   'L-BFGS-B': 'L-BFGS-B',
                   'dual_annealing': 'dual_annealing',
                   'SHGO': 'shgo',
                   'MCMC': 'MCMC'}

        # obtain optimisation parameters (maxiter, etc)
        kws = self.optimisation_parameters.parameters(alg)

        if methods[alg] != 'MCMC':
            fitter = CurveFitter(objective)
            progress = ProgressCallback(self, objective=objective)

            if alg == 'L-BFGS-B':
                maxiter = kws.pop('maxiter')
                kws['options'] = {'maxiter': maxiter}

            if alg != 'LM':
                progress.show()
                kws['callback'] = progress.callback

            try:
                # workers is added to differential evolution in scipy 1.2
                if alg == 'DE':
                    with MapWrapper(-1) as workers:
                        kws['workers'] = workers
                        fitter.fit(method=methods[alg], **kws)
                else:
                    fitter.fit(method=methods[alg], **kws)

                print(str(objective))
            except StopIteration as e:
                # user probably aborted the fit
                # but it's still worth creating a fit curve, so don't return
                # in this catch block
                # xk should be the best fit so far.
                text = e.args[0]
                xk = e.args[1]

                msg(repr(text))
                print(repr(text))

                objective.setp(xk)
                print(objective)
            except Exception as e:
                # Typically shown when sensible limits weren't provided
                msg(repr(e))
                progress.close()
                return []

            progress.close()
        else:
            dialog = SampleMCMCDialog(parent=self)
            if not dialog.exec_():
                return []

            folder_dialog = QtWidgets.QFileDialog(
                parent=self,
                caption='Select location to save MCMC output')
            folder_dialog.setFileMode(QtWidgets.QFileDialog.Directory)
            folder_dialog.setWindowModality(QtCore.Qt.WindowModal)
            if folder_dialog.exec_():
                folder = folder_dialog.selectedFiles()[0]
            else:
                return []

            nwalkers = dialog.walkers.value()
            init = dialog.init.currentText()
            nsteps = dialog.steps.value()
            nthin = dialog.thin.value()
            ntemps = dialog.temps.value()
            if ntemps in [-1, 0, 1]:
                ntemps = -1
            dialog.close()

            fitter = CurveFitter(objective, ntemps=ntemps, nwalkers=nwalkers)

            try:
                fitter.initialise(pos=init)
                progress = QtWidgets.QProgressDialog('MCMC progress',
                                                     'Abort',
                                                     0,
                                                     nsteps,
                                                     parent=self)
                progress.setWindowModality(QtCore.Qt.WindowModal)
                progress.setAutoClose(True)
                progress.setValue(0)
                progress.show()

                # only want to update every few seconds - updating the progress
                # bar can chew processor cycles.
                completed = [0]
                last_time = [time.time()]

                def callback(coords, logprob):
                    completed[0] += 1
                    if (time.time() - last_time[0]) > 2:
                        last_time[0] = time.time()
                        progress.setValue(completed[0])
                        if progress.wasCanceled():
                            raise StopIteration("Sampling aborted")

                with open(os.path.join(folder, 'steps.chain'), 'w') as f,\
                        get_context('spawn').Pool() as workers:
                    fitter.sample(nsteps, f=f, verbose=True, nthin=nthin,
                                  callback=callback, pool=workers.map)

            except StopIteration:
                pass
            except Exception as e:
                progress.close()
                msg(repr(e))
                print(repr(e))
                return []
            progress.close()

            # process the samples
            try:
                dialog = ProcessMCMCDialog(objective, fitter.chain,
                                           folder=folder, parent=self)
                dialog.exec_()
                dialog.close()
                # create MCMC graphs
                _plots(objective, nplot=dialog.nplot.value(), folder=folder)
            except Exception as e:
                dialog.close()
                msg(repr(e))
                print(repr(e))
                return []

            print(str(objective))

        return data_objects

    def update_gui_fitted_data_objects(self, data_objects):
        """
        Updates the GUI for a sequence of successfully fitted DataObject.

        The data tree is updated first with new model values, then chi2 is
        updated for all the datasets. Finally the graphs for the datasets
        are updated.
        This method is separated out from `update_gui_model`, as it is changing
        values in the tree model.

        Parameters
        ----------
        data_objects: list of DataObject
        """

        # mark models as having been updated.
        # prevent the GUI from updating whilst we change all the values.
        # This means we have to update the graphs separately at the end of
        # this method.
        self._hold_updating = True

        for data_object in data_objects:
            # retrive data_object node from the treeModel
            node = self.treeModel.data_object_node(data_object.name)
            descendants = node.descendants()
            # filter out all the parameter nodes
            par_nodes = [n for n in descendants if isinstance(n, ParNode)]
            var_par_nodes = [n for n in par_nodes if n.parameter.vary]

            for par in var_par_nodes:
                # for some reason the view doesn't know how to display float64
                # coerce to a float
                if par.parameter.stderr is not None:
                    par.parameter.stderr = float(par.parameter.stderr)
                    parent, row = par.parent(), par.row()
                    idx1 = self.treeModel.index(row, 1, parent.index)
                    idx2 = self.treeModel.index(row, 2, parent.index)
                    self.treeModel.dataChanged.emit(idx1, idx2)

        # re-enable the GUI updating whilst we change all the values
        self._hold_updating = False

        self.add_data_objects_to_graphs(data_objects)
        # calculates chi2 and redraws generative model
        self.update_gui_model(data_objects)

    @QtCore.pyqtSlot(int)
    def on_only_fitted_stateChanged(self, arg_1):
        """
        only display fitted parameters
        """
        self.treeFilter.only_fitted = arg_1
        self.treeFilter._fitted_datasets = (
            self.currently_fitting_model.datasets)
        self.treeFilter.invalidateFilter()

    @QtCore.pyqtSlot(int)
    def on_use_errors_checkbox_stateChanged(self, arg_1):
        """
        want to weight by error bars, recalculate chi2
        """

        if arg_1:
            use = True
        else:
            use = False

        self.settings.useerrors = use
        self.update_gui_model([])

    def mapToSource(self, index):
        # converts a model index in the filter model to the original model
        return self.treeFilter.mapToSource(index)

    @QtCore.pyqtSlot(QtCore.QModelIndex)
    def on_treeView_clicked(self, index):
        index = self.mapToSource(index)
        if not index.isValid():
            return None

        item = index.internalPointer()
        if not isinstance(item, ParNode):
            return

        self.currentCell = {}

        par = item.parameter
        val = par.value

        if val < 0:
            lowlim = 2 * val
            hilim = 0
        else:
            lowlim = 0
            hilim = 2 * val

        self.currentCell['item'] = item
        self.currentCell['val'] = val
        self.currentCell['lowlim'] = lowlim
        self.currentCell['hilim'] = hilim
        self.currentCell['readyToChange'] = True

    @QtCore.pyqtSlot(int)
    def on_paramsSlider_valueChanged(self, arg_1):
        # short circuit if the treeview hasn't been clicked yet
        if not hasattr(self, 'currentCell'):
            return

        c = self.currentCell
        item = c['item']

        if not c['readyToChange']:
            return

        val = (c['lowlim'] +
               (arg_1 / 1000.) * np.fabs(c['lowlim'] - c['hilim']))

        item.parameter.value = val

        # get the index for the change
        row = item.index.row()
        # who is the parent
        parent_index = item.parent().index
        index = self.treeModel.index(row, 1, parent_index)
        self.treeModel.dataChanged.emit(index, index,
                                        [QtCore.Qt.EditRole])

    @QtCore.pyqtSlot()
    def on_paramsSlider_sliderReleased(self):
        try:
            self.currentCell['readyToChange'] = False
            self.ui.paramsSlider.setValue(499)
            item = self.currentCell['item']

            val = item.parameter.value

            if val < 0:
                low_lim = 2 * val
                hi_lim = 0
            else:
                low_lim = 0
                hi_lim = 2 * val

            self.currentCell['val'] = val
            self.currentCell['lowlim'] = low_lim
            self.currentCell['hilim'] = hi_lim
            self.currentCell['readyToChange'] = True

        except (ValueError, AttributeError, KeyError):
            return

        # for some reason linked parameters don't update well when the slider
        # moves. Setting focus to the treeview and back to the slider seems
        # to make the update happen. One could issue more dataChanged signals
        # from sliderValue changed, but I'd have to figure out all the other
        # nodes
        self.ui.treeView.setFocus(QtCore.Qt.OtherFocusReason)
        self.ui.paramsSlider.setFocus(QtCore.Qt.OtherFocusReason)

    def link_action(self):
        selected_indices = self.ui.treeView.selectedIndexes()
        par_nodes_to_link = []
        for index in selected_indices:
            # from filter to model
            index = self.mapToSource(index)
            item = index.internalPointer()
            if isinstance(item, ParNode):
                par_nodes_to_link.append(item)

        # get unique nodes
        par_nodes_to_link = list(unique(par_nodes_to_link))

        if len(par_nodes_to_link) < 2:
            return

        mpn = par_nodes_to_link[0]
        mp = mpn.parameter

        # if the master parameter already has a constraint, then it
        # has to be removed, otherwise recursion occurs
        if mp.constraint is not None:
            mp.constraint = None
            idx = self.treeModel.index(mpn.row(), 1, mpn.parent().index)
            idx1 = self.treeModel.index(mpn.row(), 5, mpn.parent().index)
            self.treeModel.dataChanged.emit(idx, idx1)

        for par in par_nodes_to_link[1:]:
            par.parameter.constraint = mp
            idx = self.treeModel.index(par.row(), 1, par.parent().index)
            idx1 = self.treeModel.index(par.row(), 5, par.parent().index)
            self.treeModel.dataChanged.emit(idx, idx1)

        self.ui.paramsSlider.setFocus(QtCore.Qt.OtherFocusReason)
        self.ui.treeView.setFocus(QtCore.Qt.OtherFocusReason)

    def link_equivalent_action(self):
        # link equivalent parameters across a whole range of datasets.
        # the datasets all need to have the same structure for this to work.

        # retrieve data_objects that need to be linked
        datastore = self.treeModel.datastore

        self.data_object_selector.setWindowTitle("Select equivalent datasets"
                                                 " to link")
        ok = self.data_object_selector.exec_()
        if not ok:
            return
        items = self.data_object_selector.data_objects.selectedItems()
        names = [item.text() for item in items]
        if 'theoretical' in names:
            names.pop(names.index('theoretical'))

        # these are the data objects that you want to link across
        data_objects = [datastore[name] for name in names]

        # these are the nodes selected in the tree view. We're going to link
        # these nodes together, to equivalent nodes on the other data objects
        selected_indices = self.ui.treeView.selectedIndexes()
        par_nodes_to_link = []
        for index in selected_indices:
            # from filter to model
            index = self.mapToSource(index)
            item = index.internalPointer()
            if isinstance(item, ParNode):
                par_nodes_to_link.append(item)

        # get unique nodes
        par_nodes_to_link = list(unique(par_nodes_to_link))

        def is_same_structure(objs):
            # Now check that the models are roughly the same
            models = [data_object.model for data_object in objs]
            structures = [m.structure for m in models]
            ncomponents = [len(s) for s in structures]
            parameters = [list(flatten(m.parameters)) for m in models]
            nparams = [len(p) for p in parameters]

            if len(set(ncomponents)) == 1 and len(set(nparams)) == 1:
                return True
            return False

        extra_pars = []
        # retrieve equivalent parameters on other datasets
        for node in par_nodes_to_link:
            mstr_obj_node = find_data_object(node.index)
            # check that all data_objects have the same structure as the
            # selected parameter
            if not is_same_structure([mstr_obj_node.data_object] +
                                     data_objects):
                return msg("All models must have equivalent structural"
                           " components and the same number of parameters for"
                           " equivalent linking to be available, no linking"
                           " has been done.")

            row_indices = node.row_indices()

            for data_object in data_objects:
                # retrieve the similar parameter from row indices
                row_indices[1] = self.treeModel.data_object_row(
                    data_object.name)
                # now identify where the similar parameter is
                pn = self.treeModel.node_from_row_indices(row_indices)
                extra_pars.append(pn)

        par_nodes_to_link.extend(extra_pars)
        par_nodes_to_link = list(unique(par_nodes_to_link))

        mpn = par_nodes_to_link[0]
        mp = mpn.parameter

        # if the master parameter already has a constraint, then it
        # has to be removed, otherwise recursion occurs
        if mp.constraint is not None:
            mp.constraint = None
            idx = self.treeModel.index(mpn.row(), 1, mpn.parent().index)
            idx1 = self.treeModel.index(mpn.row(), 5, mpn.parent().index)
            self.treeModel.dataChanged.emit(idx, idx1)

        for par in par_nodes_to_link[1:]:
            par.parameter.constraint = mp
            idx = self.treeModel.index(par.row(), 1, par.parent().index)
            idx1 = self.treeModel.index(par.row(), 5, par.parent().index)
            self.treeModel.dataChanged.emit(idx, idx1)

        self.ui.paramsSlider.setFocus(QtCore.Qt.OtherFocusReason)
        self.ui.treeView.setFocus(QtCore.Qt.OtherFocusReason)

    def unlink_action(self):
        selected_indices = self.ui.treeView.selectedIndexes()
        par_nodes_to_unlink = []
        for index in selected_indices:
            # from filter to model
            index = self.mapToSource(index)
            item = index.internalPointer()
            if isinstance(item, ParNode):
                par_nodes_to_unlink.append(item)

        # get unique nodes
        par_nodes_to_unlink = list(unique(par_nodes_to_unlink))

        for par in par_nodes_to_unlink:
            par.parameter.constraint = None
            idx = self.treeModel.index(par.row(), 2, par.parent().index)
            idx1 = self.treeModel.index(par.row(), 5, par.parent().index)
            self.treeModel.dataChanged.emit(idx, idx1)

        # a trick to make the treeView repaint
        self.ui.paramsSlider.setFocus(QtCore.Qt.OtherFocusReason)
        self.ui.treeView.setFocus(QtCore.Qt.OtherFocusReason)

    def copy_from_action(self):
        # whose model did you want to use?
        datastore = self.treeModel.datastore

        model_names = datastore.names
        which_model, ok = QtWidgets.QInputDialog.getItem(
            self, "Which model did you want to copy?",
            "model",
            model_names,
            editable=False)

        if not ok:
            return

        selected_indices = self.ui.treeView.selectedIndexes()
        source_model = datastore[which_model].model

        nodes = []
        for index in selected_indices:
            # from filter to model
            index = self.mapToSource(index)
            item = index.internalPointer()
            nodes.append(item.hierarchy())

        # filter out the data_object_nodes
        data_object_nodes = [n for n in flatten(nodes) if
                             isinstance(n, DataObjectNode)]

        # get unique data object nodes, these are the data objects whose
        # model you want to overwrite
        data_object_nodes = list(unique(data_object_nodes))

        # now set the model for all those data object nodes
        # _hold_updating is set so that if a lot of things change in the
        # tree structure, the recalculation of reflectivity curves triggered
        # by set_reflect_model doesn't cause an enormous slowdown. It should
        # be sufficient to do a single update at the end.
        self._hold_updating = True
        try:
            for don in data_object_nodes:
                new_model = deepcopy(source_model)
                new_model.name = don.data_object.name
                don.set_reflect_model(new_model)
        finally:
            do = [node.data_object for node in data_object_nodes]

            self._hold_updating = False
            self.update_gui_model(do)

    @QtCore.pyqtSlot()
    def add_mixed_area_action(self):
        selected_indices = self.ui.treeView.selectedIndexes()

        if not selected_indices:
            return

        index = selected_indices[0]

        # from filter to model
        index = self.mapToSource(index)
        if not index.isValid():
            return

        item = index.internalPointer()
        data_object_node = find_data_object(index)
        reflect_model_node = data_object_node.child(1)
        structures = reflect_model_node.structures

        if isinstance(item, StructureNode):
            copied_structure = deepcopy(item._data)
        else:
            copied_structure = deepcopy(structures[0])

        # add it to the list of structures, at the END
        reflect_model_node.insert_structure(len(structures) + 3,
                                            copied_structure)

    @QtCore.pyqtSlot()
    def remove_mixed_area_action(self):
        selected_indices = self.ui.treeView.selectedIndexes()

        if not selected_indices:
            return

        index = selected_indices[0]

        # from filter to model
        index = self.mapToSource(index)
        if not index.isValid():
            return

        item = index.internalPointer()
        if not isinstance(item, StructureNode):
            return msg("Please select a single Structure to remove")

        data_object_node = find_data_object(index)
        reflect_model_node = data_object_node.child(1)
        if len(reflect_model_node.structures) == 1:
            return msg("Your model only contains a single Structure, removal"
                       " not possible")

        reflect_model_node.remove_structure(item.row())

    def modify_gui(self):
        """
        Only called at program initialisation. Used to add the plots
        """
        self.sldgraphs = MySLDGraphs(self.ui.sld)
        self.ui.gridLayout_4.addWidget(self.sldgraphs)

        self.reflectivitygraphs = MyReflectivityGraphs(self.ui.reflectivity)
        self.ui.gridLayout_5.addWidget(self.reflectivitygraphs)

        self.ui.gridLayout_5.addWidget(self.reflectivitygraphs.mpl_toolbar)
        self.ui.gridLayout_4.addWidget(self.sldgraphs.mpl_toolbar)

    def redraw_data_object_graphs(self, data_objects,
                                  all=False,
                                  transform=True):
        """
        Asks the graphs to be redrawn, delegating generative calculation to
        reflectivitygraphs.redraw_data_objects

        Parameters
        ----------
        data_objects: list
            all the datasets to be redrawn
        all: bool
            redraw all the datasets
        transform: bool or curvefitter.Transform
            Transform the data in the dataset for visualisation.
            False - don't transform
            True - transform the data using the program default
            refnx.objective.Transform - use a custom transform
        """
        datastore = self.treeModel.datastore
        if all:
            data_objects = [data_object for data_object in datastore]

        if callable(transform):
            t = transform
        elif transform:
            t = Transform(self.settings.transformdata)
        else:
            t = None

        self.reflectivitygraphs.redraw_data_objects(data_objects,
                                                    transform=t)
        self.sldgraphs.redraw_data_objects(data_objects)

    def tree_model_structure_changed(self, parent, first, last):
        # called when you've removed or added layers to the model
        if not parent.isValid():
            return

        node = parent.internalPointer()

        # if you're not adding/removing Components don't respond.
        # if not isinstance(node, StructureNode):
        #     return

        # find out which data_object / model we're adjusting
        hierarchy = node.hierarchy()
        for n in hierarchy:
            if isinstance(n, DataObjectNode):
                self.clear_data_object_uncertainties([n.data_object])
                self.update_gui_model([n.data_object])

    def clear_data_object_uncertainties(self, data_objects):
        if self._hold_updating:
            return

        for data_object in data_objects:
            # retrive data_object node from the treeModel
            node = self.treeModel.data_object_node(data_object.name)
            descendants = node.descendants()
            # filter out all the parameter nodes
            par_nodes = [n for n in descendants if isinstance(n, ParNode)]
            for par in par_nodes:
                p = par.parameter
                p.stderr = None

                parent, row = par.parent(), par.row()
                idx1 = self.treeModel.index(row, 2, parent.index)
                idx2 = self.treeModel.index(row, 2, parent.index)
                self.treeModel.dataChanged.emit(idx1, idx2)

    def tree_model_data_changed(self, top_left, bottom_right,
                                role=QtCore.Qt.EditRole):
        # if you've just changed whether you want to hold or vary a parameter
        # there is no need to update the reflectivity plots
        if not top_left.isValid():
            return

        # find out which data_object / model we're adjusting
        node = top_left.internalPointer()
        if node is None:
            return

        if (len(role) and role[0] == QtCore.Qt.CheckStateRole and
                isinstance(node, DataObjectNode)):
            # set visibility of data_object
            graph_properties = node.data_object.graph_properties
            graph_properties.visible = (node.visible is True)
            self.redraw_data_object_graphs([node.data_object])
            return

        # list of data objects to wipe and update
        wipe_update = []

        # redraw if you're altering a PropertyNode (edit or check)
        if (top_left.column() == 1 and len(role) and
            role[0] in [QtCore.Qt.CheckStateRole, QtCore.Qt.EditRole] and
                isinstance(node, PropertyNode)):
            wipe_update = [find_data_object(top_left).data_object]

        # only redraw if you're altering values
        # otherwise we'd be performing continual updates of the model
        if (top_left.column() == 1 and len(role) and
                role[0] == QtCore.Qt.EditRole and isinstance(node, ParNode)):
            param = node.parameter
            wipe_update = [find_data_object(top_left).data_object]

            # find if there are dependent parameters on the original
            # parameter. There's an argument for doing this in the treeModel,
            # the model is normally responsible for manipulating data. However,
            # if we do it here then we only need to redraw once.
            ds = self.treeModel.datastore
            for do in ds:
                model = do.model
                cpars = model.parameters.constrained_parameters()
                for cpar in cpars:
                    deps = cpar.dependencies()
                    if param in deps:
                        wipe_update.append(do)
                        break

        if wipe_update:
            wipe_update = list(unique(wipe_update))
            self.clear_data_object_uncertainties(wipe_update)
            self.update_gui_model(wipe_update)

    def calculate_chi2(self, data_objects):
        # calculate chi2 for all the data objects
        if not len(data_objects):
            return

        if self._hold_updating:
            return

        for data_object in data_objects:
            if data_object.name == 'theoretical':
                continue

            useerrors = self.settings.useerrors

            # this is if the user requested constant dq/q. If that's the case
            # then make a new dataset to hide the resolution information
            dataset_t = data_object.dataset
            if data_object.constantdq_q and dataset_t.x_err is not None:
                dataset_t = Data1D(data=dataset_t)
                dataset_t.x_err = None

            t = None
            if self.settings.transformdata is not None:
                transform = Transform(self.settings.transformdata)
                t = transform

            # can't have negative points if fitting as logR vs Q
            if self.settings.transformdata == 'logY':
                dataset_t = self.filter_neg_reflectivity(dataset_t)

            objective = Objective(data_object.model,
                                  dataset_t,
                                  transform=t,
                                  use_weights=useerrors)
            chisqr = objective.chisqr()
            node = self.treeModel.data_object_node(data_object.name)
            node.chi2 = chisqr
            index = self.treeModel.index(node.row(), 2, node.parent().index)
            self.treeModel.dataChanged.emit(index, index)

    def update_gui_model(self, data_objects):
        """
        Recalculates chi2 and the generative model for a list of DataObject
        """
        if not len(data_objects):
            return

        if self._hold_updating:
            return

        self.redraw_data_object_graphs(data_objects)
        self.calculate_chi2(data_objects)

    def filter_neg_reflectivity(self, dataset):
        # if one is fitting as logR vs Q (with/without errors), then negative
        # points mess up the transform. Therefore, filter those negative points
        copy = Data1D(data=dataset)
        copy.mask = dataset.y > 0
        return copy

    def add_data_objects_to_graphs(self, data_objects):
        transform = Transform(self.settings.transformdata)
        t = transform

        self.reflectivitygraphs.add_data_objects(data_objects, transform=t)
        self.sldgraphs.add_data_objects(data_objects)


class ProgressCallback(QtWidgets.QDialog):
    def __init__(self, parent=None, objective=None):
        self.start = time.time()
        self.last_time = time.time()
        self.abort_flag = False
        super(ProgressCallback, self).__init__(parent)
        self.parent = parent
        self.ui = uic.loadUi(os.path.join(UI_LOCATION, 'progress.ui'), self)
        self.elapsed = 0.
        self.chi2 = 1.e308
        self.ui.timer.display(float(self.elapsed))
        self.ui.buttonBox.rejected.connect(self.abort)
        self.objective = objective
        self.iterations = 0

    def abort(self):
        self.abort_flag = True

    def callback(self, xk, *args, **kwds):
        # a callback for scipy.optimize.minimize, which enters
        # every iteration.
        new_time = time.time()
        self.iterations += 1

        # update every 1.5 seconds
        if new_time - self.last_time > 1.5:
            # gp = self.dataset.graph_properties
            # if gp.line2Dfit is not None:
            #     self.parent.redraw_data_object_graphs([self.dataset])
            # else:
            #     self.parent.add_datasets_to_graphs([self.dataset])

            self.elapsed = new_time - self.start
            self.ui.timer.display(float(self.elapsed))
            self.last_time = new_time

            text = 'Chi2 : {}\nIterations : {}'.format(
                self.objective.chisqr(xk),
                self.iterations)

            self.ui.values.setPlainText(text)
            QtWidgets.QApplication.processEvents()
            if self.abort_flag:
                raise StopIteration("WARNING: FIT WAS TERMINATED EARLY", xk)

        return self.abort_flag


class ProgramSettings(object):

    def __init__(self, **kwds):
        _members = {'fitting_algorithm': 'DE',
                    'transformdata': 'logY',
                    'quad_order': 17,
                    'current_dataset_name': None,
                    'experiment_file_name': '',
                    'current_model_name': None,
                    'usedq': True,
                    'resolution': 5,
                    'fit_plugin': None,
                    'useerrors': True}

        for key in _members:
            if key in kwds:
                setattr(self, key, kwds[key])
            else:
                setattr(self, key, _members[key])

    def __getitem__(self, key):
        return self.__dict__[key]

    def __getattr_(self, key):
        return self.__dict__[key]

    def __setitem__(self, key, value):
        if key in self.__dict__:
            setattr(self, key, value)


class NavToolBar(NavigationToolbar):
    """
    This class overloads the NavigationToolbar of matplotlib.
    """
    def __init__(self, canvas, parent, coordinates=True):
        NavigationToolbar.__init__(self, canvas, parent, coordinates)
        self.setIconSize(QtCore.QSize(20, 20))


class MyReflectivityGraphs(FigureCanvas):

    """Ultimately, this is a QWidget (as well as a FigureCanvasAgg, etc.)."""

    def __init__(self, parent=None):
        self.figure = Figure(facecolor=(1, 1, 1), edgecolor=(0, 0, 0))

        # reflectivity graph
        self.axes = []
        ax = self.figure.add_axes([0.06, 0.15, 0.9, 0.8])
        ax.margins(0.0005)
        self.axes.append(ax)

        self.axes[0].autoscale(axis='both', tight=False, enable=True)
        self.axes[0].set_xlabel('Q')
        self.axes[0].set_ylabel('R')
        # self.axes[0].set_yscale('log')

# residual plot
# , sharex=self.axes[0]
# ax2 = self.figure.add_axes([0.1,0.04,0.85,0.14], sharex=ax, frame_on = False)
#   self.axes.append(ax2)
#   self.axes[1].set_visible(True)
#   self.axes[1].set_ylabel('residual')

        FigureCanvas.__init__(self, self.figure)
        self.setParent(parent)
        self.figure.canvas.setFocusPolicy(QtCore.Qt.ClickFocus)

        self.mpl_toolbar = NavToolBar(self, parent)
        self.figure.canvas.mpl_connect('pick_event', self._pick_event)
        # self.figure.canvas.mpl_connect('key_press_event', self._key_press)

        self.draw()

    def _key_press(self, event):
        # auto scale
        if event.key == 'super+a':
            self.autoscale()

    def _pick_event(self, event):
        # pick event was a double click on the graph
        if (event.mouseevent.dblclick and
                event.mouseevent.button == 1):
            if isinstance(event.artist, lines.Line2D):
                self.mpl_toolbar.edit_parameters()

    def autoscale(self):
        self.axes[0].relim()
        self.axes[0].autoscale(axis='both', tight=False, enable=True)
        self.draw()

    def add_data_objects(self, data_objects, transform=None):
        for data_object in data_objects:
            dataset = data_object.dataset

            graph_properties = data_object.graph_properties

            if (graph_properties.ax_data is None and
                    data_object.name != 'theoretical'):
                yt = dataset.y
                if transform is not None:
                    yt, edata = transform(dataset.x, dataset.y, dataset.y_err)

                # add the dataset
                line_instance = self.axes[0].plot(dataset.x,
                                                  yt,
                                                  markersize=3,
                                                  marker='o',
                                                  linestyle='',
                                                  markeredgecolor=None,
                                                  label=dataset.name,
                                                  picker=5)
                mfc = artist.getp(line_instance[0], 'markerfacecolor')
                artist.setp(line_instance[0], **{'markeredgecolor': mfc})

                graph_properties['ax_data'] = line_instance[0]
                if graph_properties['data_properties']:
                    artist.setp(graph_properties.ax_data,
                                **graph_properties['data_properties'])

            yfit_t = data_object.generative
            if graph_properties.ax_fit is None and yfit_t is not None:
                if transform is not None:
                    yfit_t, temp = transform(dataset.x, yfit_t)

                color = 'b'
                if graph_properties.ax_data is not None:
                    color = artist.getp(graph_properties.ax_data, 'color')
                # add the fit
                graph_properties['ax_fit'] = self.axes[0].plot(
                    dataset.x,
                    yfit_t,
                    linestyle='-',
                    color=color,
                    lw=1,
                    label='fit_' + data_object.name,
                    picker=5)[0]
                if graph_properties['fit_properties']:
                    artist.setp(graph_properties.ax_fit,
                                **graph_properties['fit_properties'])

            # if (dataObject.line2Dresiduals is None and
            #     dataObject.residuals is not None):
            #     dataObject.line2Dresiduals = self.axes[1].plot(
            #         dataObject.x,
            #         dataObject.residuals,
            #         linestyle='-',
            #         lw = 1,
            #         label = 'residuals_' + dataObject.name)[0]
            #
            #     if graph_properties['residuals_properties']:
            #         artist.setp(dataObject.ax_residuals,
            #                     **graph_properties['residuals_properties'])

            graph_properties.save_graph_properties()
        self.draw()

    def redraw_data_objects(self, data_objects, transform=None):
        if not len(data_objects):
            return

        for data_object in data_objects:
            if not data_object:
                continue

            dataset = data_object.dataset

            if data_object.name != 'theoretical':
                y = dataset.y
                if transform is not None:
                    y, _ = transform(dataset.x, y)

            if data_object.model is not None:
                yfit = data_object.generative

                if transform is not None:
                    yfit, efit = transform(dataset.x, yfit)

            graph_properties = data_object.graph_properties
            visible = graph_properties.visible

            if graph_properties.ax_data is not None:
                graph_properties.ax_data.set_data(dataset.x, y)
                graph_properties.ax_data.set_visible(visible)
            if graph_properties.ax_fit is not None:
                graph_properties.ax_fit.set_data(dataset.x, yfit)
                graph_properties.ax_fit.set_visible(visible)

#          if dataObject.line2Dresiduals:
#             dataObject.line2Dresiduals.set_data(dataObject.x,
#                                          dataObject.residuals)
#             dataObject.line2Dresiduals.set_visible(visible)

        self.draw()

    def remove_trace(self, data_object):
        graph_properties = data_object.graph_properties
        if graph_properties.ax_data is not None:
            graph_properties.ax_data.remove()
            graph_properties.ax_data = None

        if graph_properties.ax_fit is not None:
            graph_properties.ax_fit.remove()
            graph_properties.ax_fit = None

        if graph_properties.ax_residuals is not None:
            graph_properties.ax_residuals.remove()
            graph_properties.ax_residuals = None
        self.draw()

    def remove_traces(self):
        while len(self.axes[0].lines):
            del self.axes[0].lines[0]
        self.draw()


class MySLDGraphs(FigureCanvas):

    """Ultimately, this is a QWidget (as well as a FigureCanvasAgg, etc.)."""

    def __init__(self, parent=None):
        self.figure = Figure(facecolor=(1, 1, 1), edgecolor=(0, 0, 0))
        self.axes = []
        # SLD plot
        self.axes.append(self.figure.add_subplot(111))

        self.axes[0].autoscale(axis='both', tight=False, enable=True)
        self.axes[0].set_xlabel('z')
        self.axes[0].set_ylabel('SLD')

        FigureCanvas.__init__(self, self.figure)
        self.setParent(parent)
        self.figure.subplots_adjust(left=0.1, right=0.95, top=0.98)
        self.mpl_toolbar = NavigationToolbar(self, parent)

    def redraw_data_objects(self, data_objects):
        for data_object in data_objects:
            if not data_object:
                continue
            graph_properties = data_object.graph_properties
            visible = graph_properties.visible

            if (graph_properties.ax_sld_profile and
                    data_object.model is not None):
                try:
                    sld_profile = data_object.model.structure.sld_profile()

                    graph_properties.ax_sld_profile.set_data(
                        sld_profile[0],
                        sld_profile[1])
                    graph_properties.ax_sld_profile.set_visible(visible)
                except AttributeError:
                    # TODO, fix this
                    # this may happen for MixedReflectModel, the model doesnt
                    # have structure.sld_profile()
                    continue

        self.axes[0].relim()
        self.axes[0].autoscale_view(None, True, True)
        self.draw()

    def add_data_objects(self, data_objects):
        for data_object in data_objects:
            graph_properties = data_object.graph_properties
            if (graph_properties.ax_sld_profile is None and
                    data_object.sld_profile is not None):

                color = 'r'
                lw = 2
                if graph_properties.ax_data:
                    color = artist.getp(graph_properties.ax_data, 'color')
                    lw = artist.getp(graph_properties.ax_data, 'lw')

                try:
                    graph_properties['ax_sld_profile'] = self.axes[0].plot(
                        data_object.sld_profile[0],
                        data_object.sld_profile[1],
                        linestyle='-',
                        color=color,
                        lw=lw,
                        label='sld_' + data_object.name)[0]
                except AttributeError:
                    # this may happen for MixedReflectModel, the model doesnt
                    # have structure.sld_profile()
                    continue

                if graph_properties['sld_profile_properties']:
                    artist.setp(graph_properties.ax_sld_profile,
                                **graph_properties['sld_profile_properties'])

        self.axes[0].relim()
        self.axes[0].autoscale(axis='both', tight=False, enable=True)
        self.draw()

    def remove_trace(self, data_object):
        if data_object.graph_properties.ax_sld_profile:
            data_object.graph_properties.ax_sld_profile.remove()
        self.draw()

    def remove_traces(self):
        while len(self.axes[0].lines):
            del self.axes[0].lines[0]

        self.draw()


class EmittingStream(QtCore.QObject):
    # a class for rewriting stdout to a console window

    textWritten = QtCore.pyqtSignal(str)

    def __init__(self):
        QtCore.QObject.__init__(self)
        # super(EmittingStream, self).__init__()

    def write(self, text):
        self.textWritten.emit(str(text))

    def flush(self):
        pass


class OpenMenu(QtWidgets.QMenu):
    """
    A context menu class for the model tree view
    """
    def __init__(self, parent=None):
        super(OpenMenu, self).__init__(parent)
        self._parent = parent
        self.copy_from_action = self.addAction("Copy a model to here")
        self.addSeparator()
        self.add_to_fit_action = self.addAction("Add to fit")
        self.remove_from_fit_action = self.addAction("Remove from fit")
        self.addSeparator()
        self.link_action = self.addAction("Link parameters")
        self.unlink_action = self.addAction("Unlink parameters")
        self.link_equivalent_action = self.addAction(
            "Link equivalent parameters on other datasets")
        self.addSeparator()
        self.add_mixed_area = self.addAction('Mixed area - add a structure')
        self.remove_mixed_area = self.addAction("Mixed area - remove a"
                                                " structure")

    def __call__(self, position):
        action = self.exec_(self._parent.mapToGlobal(position))
        if action == self.link_action:
            pass
        if action == self.unlink_action:
            pass


def msg(text):
    # utility function for displaying a message
    msgBox = QtWidgets.QMessageBox()
    msgBox.setText(text)
    return msgBox.exec_()


def _default_slab(parent=None):
    # a default slab to add to a model
    material = SLD(3.47)
    c = material(15, 3)
    c.name = 'slab'
    c.thick.name = 'thick'
    c.rough.name = 'rough'
    c.sld.real.name = 'sld'
    c.sld.imag.name = 'isld'
    c.vfsolv.name = 'vfsolv'
    return c


_DataObjectDialog = uic.loadUiType(os.path.join(UI_LOCATION,
                                                'data_object_selector.ui'))[0]


class DataObjectSelectorDialog(QtWidgets.QDialog, _DataObjectDialog):
    def __init__(self, parent=None):
        # persistent data object selector dlg
        QtWidgets.QDialog.__init__(self, parent)
        self.setupUi(self)

    def addItems(self, items):
        self.data_objects.addItems(items)

    def removeItem(self, item):
        list_items = self.data_objects.findItems(item, QtCore.Qt.MatchExactly)
        if list_items:
            row = self.data_objects.row(list_items[0])
            self.data_objects.takeItem(row)


class CurrentlyFitting(QtCore.QAbstractListModel):
    """
    Keeps a list of the datasets that are currently being fitted
    """
    def __init__(self, parent=None):
        super(CurrentlyFitting, self).__init__(parent=parent)
        self.datasets = []

    def rowCount(self, index):
        return len(self.datasets)

    def data(self, index, role):
        if not index.isValid():
            return None

        if role == QtCore.Qt.DisplayRole:
            row = index.row()
            return self.datasets[row]

    def addItems(self, items):
        new_items = [i for i in items if
                     (i not in self.datasets) and (i != 'theoretical')]

        n_current = len(self.datasets)
        if new_items:
            self.beginInsertRows(QtCore.QModelIndex(),
                                 n_current,
                                 n_current + len(new_items) + 1)
            self.datasets.extend(new_items)
            self.endInsertRows()

    def removeItems(self, indices):
        # remove by number
        indices.sort()
        indices.reverse()

        for i in indices:
            self.beginRemoveRows(QtCore.QModelIndex(), i, i)
            self.datasets.pop(i)
            self.endRemoveRows()

    def flags(self, column):
        flags = QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled

        # say that you want the Components to be draggable
        flags |= QtCore.Qt.ItemIsDropEnabled

        return flags

    def supportedDropActions(self):
        return QtCore.Qt.MoveAction

    def mimeTypes(self):
        return ['application/vnd.treeviewdragdrop.list']

    def dropMimeData(self, data, action, row, column, parent):
        # drop datasets from the treeview
        if action == QtCore.Qt.IgnoreAction:
            return True
        if not data.hasFormat('application/vnd.treeviewdragdrop.list'):
            return False
        # what was dropped?
        ba = data.data('application/vnd.treeviewdragdrop.list')
        index_info = pickle.loads(ba)

        to_add = []
        for i in index_info:
            if i['name'] is not None:
                to_add.append(i['name'])

        if to_add:
            self.addItems(unique(to_add))
            return True

        return False



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = [
    "Backend",
    "HDFBackend", "TempHDFBackend",
    "get_test_backends",
]

from .backend import Backend
from .hdf import HDFBackend, TempHDFBackend


def get_test_backends():
    backends = [Backend]

    try:
        import h5py  # NOQA
    except ImportError:
        pass
    else:
        backends.append(TempHDFBackend)

    return backends



# -*- coding: utf-8 -*-

from __future__ import division, print_function

import numpy as np
from .red_blue import RedBlueMove

__all__ = ["DEMove"]


class DEMove(RedBlueMove):
    """A proposal using differential evolution.

    This `Differential evolution proposal
    <http://www.stat.columbia.edu/~gelman/stuff_for_blog/cajo.pdf>`_ is
    implemented following `Nelson et al. (2013)
    <http://arxiv.org/abs/1311.5229>`_.

    Args:
        sigma (float): The standard deviation of the Gaussian used to stretch
            the proposal vector.
        gamma0 (Optional[float]): The mean stretch factor for the proposal
            vector. By default, it is :math:`2.38 / \sqrt{2\,\mathrm{ndim}}`
            as recommended by the two references.

    """
    def __init__(self, sigma, gamma0=None, **kwargs):
        self.sigma = sigma
        self.gamma0 = gamma0
        kwargs["nsplits"] = 3
        super(DEMove, self).__init__(**kwargs)

    def setup(self, coords):
        self.g0 = self.gamma0
        if self.g0 is None:
            # Pure MAGIC:
            ndim = coords.shape[1]
            self.g0 = 2.38 / np.sqrt(2 * ndim)

    def get_proposal(self, s, c, random):
        Ns = len(s)
        Nc = list(map(len, c))
        ndim = s.shape[1]
        q = np.empty((Ns, ndim), dtype=np.float64)
        f = random.randn(Ns)
        for i in range(Ns):
            w = np.array([c[j][random.randint(Nc[j])] for j in range(2)])
            random.shuffle(w)
            g = np.diff(w, axis=0) * (1 + self.g0 * f[i])
            q[i] = s[i] + g
        return q, np.zeros(Ns, dtype=np.float64)



#!/usr/bin/python
"""
Functions for various calculations related to reflectometry
"""
import numpy as np
from scipy import stats, integrate, constants, optimize

# h / m = 3956
K = constants.h / constants.m_n * 1.e10


def div(d1, d2, L12=2859):
    """
    Calculate the angular resolution for a set of collimation conditions

    Parameters
    ----------
    d1 : float
        slit 1 opening
    d2 : float
        slit 2 opening
    L12 : float
        distance between slits

    Returns
    -------
    dtheta, alpha, beta
        dtheta is the FWHM of the Gaussian approximation to the trapezoidal
        resolution function
        alpha is the angular divergence of the penumbra
        beta is the angular divergence of the umbra

    When calculating dtheta / theta values for resolution, then dtheta is the
    value you need to use.
    See equations 11-14 in [1]_.

    .. [1] de Haan, V.-O.; de Blois, J.; van der Ende, P.; Fredrikze, H.;
    van der Graaf, A.; Schipper, M.; van Well, A. A. & J., v. d. Z. ROG, the
    neutron reflectometer at IRI Delft Nuclear Instruments and Methods in
    Physics Research A, 1995, 362, 434-453
    """
    divergence = 0.68 * 0.68 * (d1 ** 2 + d2 ** 2) / (L12 ** 2)
    alpha = (d1 + d2) / 2. / L12
    beta = abs(d1 - d2) / 2. / L12
    return np.degrees(np.sqrt(divergence)), np.degrees(alpha), np.degrees(beta)


def q(angle, wavelength):
    """
    Calculate Q given angle of incidence and wavelength

    Parameters
    ----------
    angle: float
        angle of incidence (degrees)
    wavelength: float
        wavelength of radiation (Angstrom)

    Returns
    -------
    Q: float
        Momentum transfer (A**-1)
    """
    return 4 * np.pi * np.sin(np.radians(angle)) / wavelength


def q2(omega, twotheta, phi, wavelength):
    """
    Convert angles and wavelength (lambda) to Q vector.

    Parameters
    ----------
    omega: float
        angle of incidence of beam (with respect to xy plane).
    twotheta: float
        angle between direct beam and the projection of the reflected beam onto
        xz plane.
    phi: float
        azimuthal angle between reflected beam and xz plane.
    wavelength: float

    Returns
    -------
    Qx, Qy, Qz : float, float, float
        Momentum transfer.

    Notes
    -----
    All angles are assumed to be in degrees.

    coordinate system:
    The beam is incident in the xz plane.
    x - along beam direction (in small angle approximation)
    y - transverse to beam direction, in plane of sample
    z - normal to sample plane.

    The xy plane is equivalent to the sample plane.
    """
    omega = np.radians(omega)
    twotheta = np.radians(twotheta)
    phi = np.radians(phi)

    qx = (2 * np.pi / wavelength * (np.cos(twotheta - omega) * np.cos(phi) -
          np.cos(omega)))
    qy = 2 * np.pi / wavelength * np.cos(twotheta - omega) * np.sin(phi)
    qz = 2 * np.pi / wavelength * (np.sin(twotheta - omega) + np.sin(omega))

    return qx, qy, qz


def wavelength(q, angle):
    """
    calculate wavelength given Q vector and angle

    Parameters
    ----------
    q : float
        Wavevector (A**-1)
    angle : float
        angle of incidence (degrees)

    Returns
    -------
    wavelength : float
        Wavelength of radiation (A)
    """
    return 4. * np.pi * np.sin(angle * np.pi / 180.) / q


def angle(q, wavelength):
    """
    calculate angle given Q and wavelength

    Parameters
    ----------
    q : float
        Wavevector (A**-1)
    wavelength : float
        Wavelength of radiation (A)

    Returns
    -------
    angle : float
        angle of incidence (degrees)
    """
    return np.arcsin((q * wavelength) / 4. / np.pi) * 180 / np.pi


def qcrit(SLD1, SLD2):
    """
    calculate critical Q vector given SLD of super and subphases

    Parameters
    ----------
    SLD1: float
        SLD of superphase (10^-6 A^-2)
    SLD2: float
        SLD of subphase (10^-6 A^-2)

    Returns
    -------
    qc: float
        Critical Q vector for a reflectivity system.
    """
    return np.sqrt(16. * np.pi * (SLD2 - SLD1) * 1.e-6)


def tauC(wavelength, xsi=0, z0=0.358, freq=24):
    """
    Calculates the burst time of a double chopper pair

    Parameters
    ----------
    wavelength: float
        wavelength in Angstroms
    z0: float
        distance between chopper pair (m)
    freq: float
        rotation frequency of choppers (Hz)
    xsi: float
        phase opening of chopper pair (degrees)

    Returns
    -------
    tauC: float
        The burst time of a double chopper pair (s)

    References
    ----------
    [1] A. A. van Well and H. Fredrikze, On the resolution and intensity of a
    time-of-flight neutron reflectometer, Physica B 357 (2005) 204-207
    [2] A. Nelson and C. Dewhurst, Towards a detailed resolution smearing
    kernel for time-of-flight neutron reflectometers, J. Appl. Cryst. (2013)
    46, 1338-1343
    """
    tauC = z0 / wavelength_velocity(wavelength)
    tauC += np.radians(xsi) / (2 * np.pi * freq)

    return tauC


def wavelength_velocity(wavelength):
    r"""
    Converts wavelength to neutron velocity

    Parameters
    ----------
    wavelength: float
        wavelength of neutron in Angstrom.

    Returns
    -------
    velocity: float
        velocity of neutron in ms**-1
    """
    return K / wavelength


def velocity_wavelength(velocity):
    r"""
    Converts neutron velocity to wavelength

    Parameters
    ----------
    velocity: float
        velocity of neutron in m/s

    Returns
    -------
    wavelength: float
        wavelength of neutron in A
    """
    return K / velocity


def wavelength_energy(wavelength):
    r"""
    Converts wavelength to energy in meV

    Parameters
    ----------
    wavelength : float
        Wavelength in Angstrom.

    Returns
    -------
    energy : float
        Energy in meV.

    """
    c = 0.5e23 / constants.eV * constants.h ** 2 / constants.m_n
    return c / wavelength ** 2


def energy_wavelength(energy):
    r"""
    Converts wavelength to energy in meV

    Parameters
    ----------
    energy : float
        Energy in meV.

    Returns
    -------
    wavelength : float
        Wavelength in Angstrom.
    """
    c = 0.5e23 / constants.eV * constants.h ** 2 / constants.m_n
    return np.sqrt(c / energy)


def double_chopper_frequency(min_wavelength, max_wavelength, L, N=1):
    r"""
    Calculates the maximum frequency available for a given wavelength band
    without getting frame overlap in a chopper spectrometer.

    Parameters
    ----------
    min_wavelength: float
        minimum wavelength to be used
    max_wavelength: float
        maximum wavelength to be used
    L: float
        Flight length of instrument (m)
    N: float, optional
        number of windows in chopper pair

    Returns
    -------
    max_freq: float
        The maximum frequency a double chopper system can use to avoid frame
        overlap.
    """
    return K / ((max_wavelength - min_wavelength) * L * N)


def resolution_double_chopper(wavelength, z0=0.358, R=0.35, freq=24,
                              H=0.005, xsi=0, L=7.5, tau_da=0):
    """
    Calculates the fractional resolution of a double chopper pair, dl/l.

    Parameters
    ----------
    wavelength: float
        wavelength in Angstroms
    z0: float
        distance between chopper pair (m)
    R: float
        radius of chopper discs (m)
    freq: float
        rotation frequency of choppers (Hz)
    N: float
        number of windows in chopper pair
    H: float
        height of beam (m)
    xsi: float
        phase opening of chopper pair (degrees)
    L: float
        Flight length of instrument (m)
    tau_da : float
        Width of timebin (s)

    Returns
    -------
    res: float
        Fractional wavelength resolution of a double chopper system.
    """
    TOF = L / wavelength_velocity(wavelength)
    tc = tauC(wavelength, xsi=xsi, z0=z0, freq=freq)
    tau_h = H / R / (2 * np.pi * freq)
    return 0.68 * np.sqrt((tc / TOF)**2 + (tau_h / TOF)**2 + (tau_da / TOF)**2)


def resolution_single_chopper(wavelength, R=0.35, freq=24, H=0.005, phi=60,
                              L=7.5):
    """
    Calculates the fractional resolution of a single chopper, dl/l.

    Parameters
    ----------
    wavelength: float
        wavelength in Angstroms
    R: float
        radius of chopper discs (m)
    freq: float
        rotation frequency of choppers (Hz)
    N: float
        number of windows in chopper
    H: float
        height of beam (m)
    phi: float
        angular opening of chopper window (degrees)
    L: float
        Flight length of instrument (m)

    Returns
    -------
    transmission: float
        Transmission of a single chopper system.
    """
    TOF = L / wavelength_velocity(wavelength)
    tauH = H / R / (2. * np.pi * freq)
    tauC = np.radians(phi) / (2. * np.pi * freq)
    return 0.68 * np.sqrt((tauC / TOF)**2 + (tauH / TOF)**2)


def transmission_double_chopper(wavelength, z0=0.358, R=0.35, freq=24, N=1,
                                H=0.005, xsi=0):
    """
    Calculates the transmission of a double chopper pair

    Parameters
    ----------
    wavelength: float
        wavelength in Angstroms
    z0: float
        distance between chopper pair (m)
    R: float
        radius of chopper discs (m)
    freq: float
        rotation frequency of choppers (Hz)
    N: float
        number of windows in chopper pair
    H: float
        height of beam (m)
    xsi: float
        phase opening of chopper pair (degrees)

    Returns
    -------
    transmission: float
        The transmission of a double chopper system.

    References
    ----------
    [1] A. A. van Well and H. Fredrikze, On the resolution and intensity of a
    time-of-flight neutron reflectometer, Physica B 357 (2005) 204-207
    [2] A. Nelson and C. Dewhurst, Towards a detailed resolution smearing
    kernel for time-of-flight neutron reflectometers, J. Appl. Cryst. (2013)
    46, 1338-1343
    """
    transmission = tauC(wavelength, xsi=xsi, z0=z0, freq=freq) * freq
    transmission += H * constants.h / (2 * np.pi * R)
    return transmission * N


def transmission_single_chopper(R=0.35, phi=60, N=1, H=0.005):
    """
    Calculates the transmission of a single chopper

    Parameters
    ----------
    R: float
        radius of chopper discs (m)
    phi: float
        angular opening of chopper disc (degrees)
    N: float
        number of windows in chopper pair
    H: float
        height of beam (m)

    References
    ----------
    [1] A. A. van Well and H. Fredrikze, On the resolution and intensity of a
    time-of-flight neutron reflectometer, Physica B 357 (2005) 204-207
    [2] A. Nelson and C. Dewhurst, Towards a detailed resolution smearing
    kernel for time-of-flight neutron reflectometers, J. Appl. Cryst. (2013)
    46, 1338-1343
    """
    return N * (np.radians(phi) * R + H) / (2 * np.pi * R)


def xray_wavelength(energy):
    """
    convert energy (keV) to wavelength (angstrom)
    """
    return 12.398 / energy


def xray_energy(wavelength):
    """
    convert energy (keV) to wavelength (angstrom)
    """
    return 12.398 / wavelength


def penetration_depth(qq, rho):
    """
    Calculates the penetration depth of a neutron/xray beam

    Parameters
    ----------
    qq: float
        Q values to calculate the penetration depth at
    rho: float or complex
        Complex SLD of material

    Returns
    -------
    penetration_depth: float
    """
    kk = 0.25 * qq ** 2.
    kk -= 4 * np.pi * rho
    temp = np.sqrt(kk + 0j)
    return np.abs(1 / temp.imag)


def beamfrac(FWHM, length, angle):
    """
    Calculate the beam fraction intercepted by a sample.

    Parameters
    ----------
    FWHM: float
        The FWHM of the beam height
    length: float
        Length of the sample in mm
    angle: float
        Angle that the sample makes w.r.t the beam (degrees)

    Returns
    -------
    beamfrac: float
        The fraction of the beam that intercepts the sample
    """
    height_of_sample = length * np.sin(np.radians(angle))
    beam_sd = FWHM / 2 / np.sqrt(2 * np.log(2))
    probability = 2. * (stats.norm.cdf(height_of_sample / 2. / beam_sd) - 0.5)
    return probability


def beamfrackernel(kernelx, kernely, length, angle):
    """
    The beam fraction intercepted by a sample, used for calculating footprints.

    Parameters
    ----------
    kernelx: array-like
        x axis for the probability kernel
    kernely: array-like
        probability kernel describing the intensity distribution of the beam
    length: float
        length of the sample
    angle: float
        angle of incidence (degrees)

    Returns
    -------
    fraction: float
        The fraction of the beam intercepted by a sample.
    """
    height_of_sample = length * np.sin(np.radians(angle))
    total = integrate.simps(kernely, kernelx)
    lowlimit = np.where(-height_of_sample / 2. >= kernelx)[0][-1]
    hilimit = np.where(height_of_sample / 2. <= kernelx)[0][0]

    area = integrate.simps(kernely[lowlimit: hilimit + 1],
                           kernelx[lowlimit: hilimit + 1])
    return area / total


def height_of_beam_after_dx(d1, d2, L12, distance):
    """
    Calculate the total widths of beam a given distance away from a collimation
    slit.

    if distance >= 0, then it's taken to be the distance after d2.
    if distance < 0, then it's taken to be the distance before d1.

    Parameters
    ----------
    d1: float
        opening of first collimation slit
    d2: float
        opening of second collimation slit
    L12: float
        distance between first and second collimation slits
    distance: float
        distance from first or last slit to a given position

    Notes
    -----
    Units - equivalent distances (inches, mm, light years)

    Returns
    -------
    (umbra, penumbra): float, float
        full width of umbra and penumbra

    """

    alpha = (d1 + d2) / 2. / L12
    beta = abs(d1 - d2) / 2. / L12
    if distance >= 0:
        return (beta * distance * 2) + d2, (alpha * distance * 2) + d2
    else:
        return ((beta * abs(distance) * 2) + d1,
                (alpha * abs(distance) * 2) + d1)


def actual_footprint(d1, d2, L12, L2S, angle):
    """
    Calculate the actual footprint on a reflectivity sample.

    Parameters
    ----------
    d1: float
        opening of first collimation slit
    d2: float
        opening of second collimation slit
    L12: float
        distance between first and second collimation slits
    L2S: float
        distance from second collimation slit to sample
    angle: float
        angle of incidence of sample (degrees)

    Returns
    -------
    (umbra_footprint, penumbra_footprint): float, float
        Footprint of the umbra and penumbra

    """
    umbra, penumbra = height_of_beam_after_dx(d1, d2, L12, L2S)
    return umbra / np.radians(angle), penumbra / np.radians(angle)


def slit_optimiser(footprint,
                   resolution,
                   angle=1.,
                   L12=2859.5,
                   L2S=180,
                   LS3=290.5,
                   LSD=2500,
                   verbose=True):
    """
    Optimise slit settings for a given angular resolution, and a given
    footprint.

    footprint: float
        maximum footprint onto sample (mm)
    resolution: float
        fractional dtheta/theta resolution (FWHM)
    angle: float, optional
        angle of incidence in degrees
    """
    if verbose:
        print('_____________________________________________')
        print('FOOTPRINT calculator - Andrew Nelson 2013')
        print('INPUT')
        print('footprint:', footprint, 'mm')
        print('fractional angular resolution (FWHM):', resolution)
        print('theta:', angle, 'degrees')

    def d1star(d2star):
        return np.sqrt(1 - np.power(d2star, 2))

    L1star = 0.68 * footprint / L12 / resolution

    def gseekfun(d2star):
        return np.power((d2star + L2S / L12 * (d2star + d1star(d2star))) -
                        L1star, 2)

    res = optimize.minimize_scalar(gseekfun, method='bounded', bounds=(0, 1))
    if res['success'] is False:
        print('ERROR: Couldnt find optimal solution, sorry')
        return None

    optimal_d2star = res['x']
    optimal_d1star = d1star(optimal_d2star)
    if optimal_d2star > optimal_d1star:
        # you found a minimum, but this may not be the optimum size of the
        # slits.
        multfactor = 1
        optimal_d2star = 1 / np.sqrt(2)
        optimal_d1star = 1 / np.sqrt(2)
    else:
        multfactor = optimal_d2star / optimal_d1star

    d1 = optimal_d1star * resolution / 0.68 * np.radians(angle) * L12
    d2 = d1 * multfactor

    tmp, height_at_S4 = height_of_beam_after_dx(d1, d2, L12, L2S + LS3)
    tmp, height_at_detector = height_of_beam_after_dx(d1, d2, L12, L2S + LSD)
    tmp, _actual_footprint = actual_footprint(d1, d2, L12, L2S, angle)

    if verbose:
        print('OUTPUT')
        if multfactor == 1:
            print('Your desired resolution results in a smaller footprint than'
                  ' the sample supports.')
            suggested_resolution = resolution * footprint / _actual_footprint
            print('You can increase flux using a resolution of',
                  suggested_resolution, 'and still keep the same footprint.')
        print('d1', d1, 'mm')
        print('d2', d2, 'mm')
        print('footprint:', _actual_footprint, 'mm')
        print('height at S4:', height_at_S4, 'mm')
        print('height at detector:', height_at_detector, 'mm')
        print('[d2star', optimal_d2star, ']')
        print('_____________________________________________')

    return d1, d2


def _dict_compare(d1, d2):
    """
    Rudimentary check to see if two dict are the same. This won't do recursive
    checking (e.g. if items in d1 or d2 are dicts)

    Parameters
    ----------
    d1 : dict
    d2 : dict

    Returns
    -------
    truth : bool
        Are two dicts the same
    """
    if len(d1) != len(d2):
        return False

    d1_keys = set(d1.keys())
    d2_keys = set(d2.keys())
    intersect_keys = d1_keys.intersection(d2_keys)
    if len(intersect_keys) != len(d1):
        return False

    for o in intersect_keys:
        if (isinstance(d1[o], np.ndarray) and
                isinstance(d2[o], np.ndarray)):
            # both numpy arrays
            if not np.array_equal(d1[o], d2[o]):
                return False
            continue

        if not isinstance(d1[o], d2[o].__class__) or d1[o] != d2[o]:
            return False

    return True



#!/usr/bin/env python
"""
run.py [options] ASV_COMMAND..

Convenience wrapper around the ``asv`` command; just sets environment
variables and chdirs to the correct place etc.

"""
from __future__ import division, absolute_import, print_function

import os
import sys
import subprocess
import json
import shutil
import argparse
import sysconfig
import errno


EXTRA_PATH = ['']


def main():
    class ASVHelpAction(argparse.Action):
        nargs = 0
        def __call__(self, parser, namespace, values, option_string=None):
            sys.exit(run_asv(['--help']))

    p = argparse.ArgumentParser(usage=__doc__.strip())
    p.add_argument('--help-asv', nargs=0, action=ASVHelpAction,
        help="""show ASV help""")
    p.add_argument('asv_command', nargs=argparse.REMAINDER)
    args = p.parse_args()

    sys.exit(run_asv(args.asv_command))


def run_asv(args):
    cwd = os.path.abspath(os.path.dirname(__file__))

    repo_dir = os.path.join(cwd, 'refnx')

    cmd = ['asv'] + list(args)
    env = dict(os.environ)

    # Inject ccache/f90cache paths
    if sys.platform.startswith('linux'):
        env['PATH'] = os.pathsep.join(EXTRA_PATH + env.get('PATH', '').split(os.pathsep))

    # Check refnx version if in dev mode; otherwise clone and setup results
    # repository
    if args and (args[0] == 'dev' or '--python=same' in args):
        import refnx
        print("Running benchmarks for refnx version %s at %s" % (refnx.__version__, refnx.__file__))

    # Run
    try:
        return subprocess.call(cmd, env=env, cwd=cwd)
    except OSError as err:
        if err.errno == errno.ENOENT:
            print("Error when running '%s': %s\n" % (" ".join(cmd), str(err),))
            print("You need to install Airspeed Velocity https://spacetelescope.github.io/asv/")
            print("to run refnx benchmarks")
            return 1
        raise


if __name__ == "__main__":
    sys.exit(main())



from __future__ import division

import numpy as np
from numpy.testing import (assert_almost_equal, assert_equal, assert_,
                           assert_allclose)

from refnx.reflect import (SLD, Structure, Spline, Slab, LipidLeaflet)
from refnx.reflect.structure import _profile_slicer
from refnx.analysis import Parameter, Interval


class TestLipidLeaflet(object):
    def setup_method(self):
        self.b_h = 6.01e-4
        self.V_h = 319.
        self.b_t = -2.92e-4
        self.V_t = 782.
        self.APM = 60.
        self.thick_h = 9.
        self.thick_t = 14.
        self.leaflet = LipidLeaflet(self.APM,
                                    self.b_h, self.V_h, self.thick_h,
                                    self.b_t, self.V_t, self.thick_t,
                                    2, 3)

        self.rho_h = self.b_h / self.V_h / 1e-6
        self.rho_t = self.b_t / self.V_t / 1e-6
        self.phi_solv_h = 1 - (self.V_h / (self.APM * self.thick_h))
        self.phi_solv_t = 1 - (self.V_t / (self.APM * self.thick_t))

    def test_slabs(self):
        # check that slab calculation from parameters is correct
        slabs = self.leaflet.slabs()
        theoretical = np.array([[self.thick_h, self.rho_h, 0, 3,
                                 self.phi_solv_h],
                                [self.thick_t, self.rho_t, 0, 2,
                                 self.phi_solv_t]])
        assert_allclose(slabs, theoretical, rtol=1e-15)

        # check that we can flip the lipid leaflet
        self.leaflet.reverse_monolayer = True
        theoretical = np.flipud(theoretical)
        theoretical[:, 3] = theoretical[::-1, 3]
        assert_allclose(self.leaflet.slabs(), theoretical, rtol=1e-15)

    def test_solvent_penetration(self):
        # check different types of solvation for heads/tails.
        self.leaflet.head_solvent = SLD(1.23)
        slabs = self.leaflet.slabs()
        assert_allclose(slabs[0, 1], 1.23 * self.phi_solv_h +
                        (1 - self.phi_solv_h) * self.rho_h)
        assert_allclose(slabs[0, 4], 0)

        self.leaflet.head_solvent = None
        self.leaflet.tail_solvent = SLD(1.23)
        slabs = self.leaflet.slabs()
        assert_allclose(slabs[1, 1], 1.23 * self.phi_solv_t +
                        (1 - self.phi_solv_t) * self.rho_t)
        assert_allclose(slabs[1, 4], 0)

    def test_initialisation_with_SLD(self):
        # we should be able to initialise with SLD objects
        heads = SLD(6.01e-4 + 0j)
        tails = SLD(-2.92e-4 + 0j)
        new_leaflet = LipidLeaflet(self.APM,
                                   heads, self.V_h, self.thick_h,
                                   tails, self.V_t, self.thick_t,
                                   2, 3)
        slabs = self.leaflet.slabs()
        new_slabs = new_leaflet.slabs()
        assert_allclose(new_slabs, slabs)

    def test_repr(self):
        # test that we can reconstruct the object from a repr
        s = repr(self.leaflet)
        q = eval(s)
        assert_equal(q.slabs(), self.leaflet.slabs())



"""
refnx: Neutron and X-ray reflectometry analysis in Python
=========================================================
Documentation is available in the docstrings and
online at https://readthedocs.org/projects/refnx/
"""

from refnx.version import version as __version__

from refnx._lib._testutils import PytestTester


test = PytestTester(__name__)
del PytestTester



import pickle
import numpy as np
from numpy.testing import (assert_allclose, assert_equal, assert_almost_equal,
                           assert_)
from refnx.reflect import SLD, Slab, Structure, Spline, Linear
from refnx.analysis import Parameter, Interval, Parameters
from refnx._lib import flatten


class TestSpline(object):

    def setup_method(self):
        self.left = SLD(1.5)(10, 3)
        self.right = SLD(2.5)(10, 3)
        self.solvent = SLD(10)

    def test_spline_smoke(self):
        # smoke test to make Spline at least gives us something
        a = Spline(100, [2],
                   [0.5], zgrad=False, microslab_max_thickness=1)

        s = self.left | a | self.right | self.solvent
        b = a.slabs(s)
        assert_equal(b[:, 2], 0)

        # microslabs are assessed in the middle of the slab
        assert_equal(b[0, 1], a(0.5 * b[0, 0], s))

        # with the ends turned off the profile should be a straight line
        assert_equal(a(50, s), 2.0)

        # construct a structure
        a = Spline(100, [2., 3., 4.],
                   [0.25] * 3, zgrad=False, microslab_max_thickness=1)

        # s.solvent = None
        s = self.left | a | self.right | self.solvent
        # calculate an SLD profile
        s.sld_profile()
        # ask for the parameters
        for p in flatten(s.parameters):
            assert_(isinstance(p, Parameter))

        # s.solvent is not None
        s.solvent = self.solvent
        # calculate an SLD profile
        s.sld_profile()

    def test_spline_no_knots(self):
        # try and make Spline with no knots
        a = Spline(100, [], [], zgrad=False, microslab_max_thickness=1)

        s = self.left | a | self.right | self.solvent
        b = a.slabs(s)
        assert_equal(b[:, 2], 0)

        # microslabs are assessed in the middle of the slab
        assert_equal(b[0, 1], a(0.5 * b[0, 0], s))

        # with the ends turned off the profile should be a straight line
        assert_equal(a(50, s), 2.0)

        q = np.linspace(0.01, 0.5, 1001)
        s.reflectivity(q)

    def test_pickle(self):
        a = Spline(100, [2, 3],
                   [0.3, 0.3], zgrad=False, microslab_max_thickness=1)

        s = self.left | a | self.right | self.solvent

        # cause the spline to be evaluated
        s.sld_profile()
        assert(a._Spline__cached_interpolator['interp'] is not None)

        pkl = pickle.dumps(s)
        r = pickle.loads(pkl)
        assert(isinstance(r, Structure))

    def test_spline_solvation(self):
        a = Spline(100, [2],
                   [0.5], zgrad=False, microslab_max_thickness=1)

        front = SLD(0.1)
        air = SLD(0.0)
        s = front | self.left | a | self.right | self.solvent

        # assign a solvent
        s.solvent = air
        self.left.vfsolv.value = 0.5
        self.right.vfsolv.value = 0.5
        assert_equal(s.slabs()[1, 1], 0.75)
        assert_equal(s.slabs()[-2, 1], 1.25)

        assert_almost_equal(a(0, s), 0.75)
        assert_almost_equal(a(100, s), 1.25)

        # remove solvent, should be solvated by backing medium
        s.solvent = None
        assert_equal(s.slabs()[1, 1], 5.75)
        assert_equal(s.slabs()[-2, 1], 6.25)
        assert_almost_equal(a(0, s), 5.75)
        assert_almost_equal(a(100, s), 6.25)

        # reverse structure, should be solvated by fronting medium
        s.reverse_structure = True
        assert_equal(s.slabs()[1, 1], 1.3)
        assert_equal(s.slabs()[-2, 1], 0.8)
        # note that a(0, s) end becomes the end when the structure is reversed
        assert_almost_equal(a(0, s), 0.8)
        assert_almost_equal(a(100, s), 1.3)

    def test_left_right_influence(self):
        # make sure that if the left and right components change, so does the
        # spline
        a = Spline(100, [2],
                   [0.5], zgrad=False, microslab_max_thickness=1)

        s = self.left | a | self.right | self.solvent

        # change the SLD of the left component, spline should respond
        self.left.sld.real.value = 2.
        assert_almost_equal(a(0, s), 2)

        # check that the spline responds if it's a vfsolv that changes
        self.left.vfsolv.value = 0.5
        assert_almost_equal(Structure.overall_sld(self.left.slabs(),
                                                  self.solvent)[0, 1],
                            6.)
        assert_almost_equal(a(0, s), 6.)

        # check that the right side responds.
        self.right.sld.real.value = 5.
        assert_almost_equal(a(100, s), 5.)

        # the spline should respond if the knot SLD's are changed
        a.vs[0].value = 3.
        assert_almost_equal(a(50, s), 3.)

        # spline responds if the interval knot spacing is changed
        a.dz[0].value = 0.9
        assert_almost_equal(a(90, s), 3.)

    def test_repr(self):
        # make sure that if the left and right components change, so does the
        # spline
        a = Spline(100, [2, 3, 4],
                   [0.1, 0.2, 0.3], zgrad=False, microslab_max_thickness=1)

        s = self.left | a | self.right | self.solvent

        # should be able to repr the whole lot
        q = repr(s)
        r = eval(q)
        assert_equal(r.slabs(), s.slabs())

    def test_spline_repeat(self):
        # can't have two splines in a row.
        a = Spline(100, [2],
                   [0.5], zgrad=False, microslab_max_thickness=1)

        s = self.left | a | a | self.right | self.solvent

        from pytest import raises
        with raises(ValueError):
            s.slabs()

    def test_spine_interfaces(self):
        a = Spline(100, [2, 3],
                   [0.3, 0.3], zgrad=False, microslab_max_thickness=1)

        s = self.left | a | self.right | self.solvent

        # this is a check to ensure that:
        # 1) _micro_slabs is being called by Structure.slabs()
        # 2) _micro_slabs works correctly when a Spline is being used.
        #    Spline.interfaces should be `None`, which should expanded from
        #    that default in _micro_slabs to `[Erf] * len(Structure.slabs())`.
        s[-1].interfaces = Linear()
        s._micro_slabs()
        assert_equal(len(s._micro_slabs()), len(s.slabs()))

        # should be able to set all the interfaces in Spline to Linear.
        a.interfaces = Linear()
        a.interfaces = [Linear()]



from scipy.stats import rv_continuous, uniform
from scipy.stats._distn_infrastructure import rv_frozen
from scipy._lib._util import check_random_state
import numpy as np


class Bounds(object):
    """
    A base class that describes the probability distribution for a parameter

    """
    def __init__(self, seed=None):
        self._random_state = check_random_state(seed)

    def logp(self, value):
        """
        Calculate the log-prior probability of a value with the probability
        distribution.

        Parameters
        ----------
        val : float or array-like
            variates to calculate the log-probability for

        Returns
        -------
        arr : float or np.ndarray
            The log-probabilty corresponding to each of the variates

        """
        raise NotImplementedError

    def valid(self, val):
        """
        Checks whether a parameter value is within the support of the
        distribution. If it isn't then it returns a value that is within the
        support.

        Parameters
        ----------
        val : array-like
            values to examine

        Returns
        -------
        valid : array-like
            valid values within the support

        """
        raise NotImplementedError

    def rvs(self, size=1, random_state=None):
        """
        Generate random variates from the probability distribution.

        Parameters
        ----------
        size : int or tuple
            Specifies the number, or array shape, of random variates to return.
        random_state : None, int, float or np.random.RandomState
            For reproducible sampling

        Returns
        -------
        arr : array-like
            Random variates from within the probability distribution.

        """
        raise NotImplementedError


class PDF(Bounds):
    """
    A class that describes the probability distribution for a parameter.

    Parameters
    ----------
    rv : :class:`scipy.stats.rv_continuous` or Object
        A continuous probability distribution. If `rv` is not an
        `rv_continuous`, then it must implement the `logpdf` and `rvs`
        methods.
    seed : int, float or np.random.RandomState
        Seed for random variates

    Examples
    --------

    >>> import scipy.stats as stats
    >>> from refnx.analysis import Parameter, PDF
    >>> p = Parameter(0.5)
    >>> # use a normal distribution for prior, mean=5 and sd=1.
    >>> p.bounds = PDF(stats.norm(5, 1))
    >>> p.logp(), stats.norm.logpdf(0.5, 5, 1)
    (-11.043938533204672, -11.043938533204672)

    """
    def __init__(self, rv, seed=None):
        super(PDF, self).__init__(seed=seed)
        # we'll accept any object so long as it has logpdf and rvs methods
        if hasattr(rv, 'logpdf') and hasattr(rv, 'rvs'):
            self.rv = rv
        else:
            raise ValueError("You must initialise PDF with an object that has"
                             " logpdf and rvs methods")

    def __repr__(self):
        return "PDF({rv!r})".format(**self.__dict__)

    def logp(self, val):
        """
        Calculate the log-prior probability of a value with the probability
        distribution.

        Parameters
        ----------
        val : float or array-like
            variate to calculate the log-probability for

        Returns
        -------
        arr : float or np.ndarray
            The log-probabilty corresponding to each of the variates

        """
        return self.rv.logpdf(val)

    def valid(self, val):
        """
        Checks whether a parameter value is within the support of the
        distribution. If it isn't then it returns a *random* value that is
        within the support.

        Parameters
        ----------
        val : array-like
            values to examine

        Returns
        -------
        valid : array-like
            valid values within the support
        """
        _val = np.asarray(val, dtype=float)
        valid = np.where(np.isfinite(self.logp(_val)),
                         _val,
                         self.rv.rvs(size=_val.shape))

        return valid

    def rvs(self, size=1, random_state=None):
        """
        Generate random variates from the probability distribution.

        Parameters
        ----------
        size : int or tuple
            Specifies the number, or array shape, of random variates to return.
        random_state : None, int, float or np.random.RandomState
            For reproducible sampling

        Returns
        -------
        arr : array-like
            Random variates from within the probability distribution.

        """
        if random_state is None:
            random_state = self._random_state
        return self.rv.rvs(size=size, random_state=random_state)


class Interval(Bounds):
    r"""
    Describes a uniform probability distribution. May be open, semi-open,
    or closed.

    Parameters
    ----------
    lb : float
        The lower bound
    ub : float
        The upper bound
    seed : None, int, float or np.random.RandomState
        For reproducible sampling

    Examples
    --------

    >>> from refnx.analysis import Parameter, Interval
    >>> p = Parameter(1)
    >>> # closed interval
    >>> p.bounds = Interval(0, 10)
    >>> p.logp([5, -1])
    array([-2.30258509,        -inf])

    A semi-closed interval will still prevent the fitter from accessing
    impossible locations.

    >>> p.bounds = Interval(lb=-10)
    >>> p.logp([5, -1])
    array([0., 0.])

    """
    def __init__(self, lb=-np.inf, ub=np.inf, seed=None):
        super(Interval, self).__init__(seed=seed)
        if lb is None:
            lb = -np.inf
        if ub is None:
            ub = np.inf

        self._lb = lb
        self._ub = ub
        self._logprob = 0
        self.rv = None
        self._closed_bounds = False
        self._set_bounds(self._lb, self._ub)

    def _set_bounds(self, lb, ub):
        t_lb, t_ub = lb, ub
        self._lb = min(t_lb, t_ub)
        self._ub = max(t_lb, t_ub)
        self._closed_bounds = False

        if np.isnan([self._lb, self._ub]).any():
            raise ValueError("Can't set Interval with NaN")

        if np.isfinite([self._lb, self._ub]).all():
            self._closed_bounds = True
            if self._lb == self._ub:
                self._logprob = 0.
            else:
                self._logprob = np.log(1 / (self._ub - self._lb))

            self.rv = uniform(self._lb, self._ub - self._lb)
        else:
            self._logprob = 0.
            self._closed_bounds = False

    def __repr__(self):
        lb, ub = self.lb, self.ub
        if np.isneginf(self.lb):
            lb = '-np.inf'
        if np.isinf(self.ub):
            ub = 'np.inf'
        return "Interval(lb={}, ub={})".format(lb, ub)

    def __str__(self):
        return '[{0}, {1}]'.format(self.lb, self.ub)

    @property
    def lb(self):
        """
        Lower bound of uniform distribution
        """
        return self._lb

    @lb.setter
    def lb(self, val):
        if val is None:
            val = -np.inf
        self._set_bounds(val, self._ub)

    @property
    def ub(self):
        """
        Upper bound of uniform distribution
        """
        return self._ub

    @ub.setter
    def ub(self, val):
        if val is None:
            val = np.inf
        self._set_bounds(self._lb, val)

    def logp(self, val):
        """
        Calculate the log-prior probability of a value with the Interval.

        Parameters
        ----------
        val : float or array-like
            variates to calculate the log-probability for

        Returns
        -------
        arr : float or np.ndarray
            The log-probabilty corresponding to each of the variates
        """
        # Special case when val is array-like. We'll assume that it may be
        # array-like if it has a length. This special casing is done because
        # a lot of time is spent on the `asarray`, `logical_and` and `where`.
        # Most of the time val is a single float, so this speedup is
        # appreciable.
        if hasattr(val, '__len__'):
            _val = np.asarray(val, dtype=float)
            valid = np.logical_and(self._lb <= _val, _val <= self._ub)
            prob = np.where(valid, self._logprob, -np.inf)
            return prob

        if self._lb <= val <= self._ub:
            return self._logprob

        return -np.inf

    def valid(self, val):
        """
        Checks whether a parameter value is within the support of the
        Interval. If it isn't then it returns a value that is within the
        support.
        If the Interval is closed (i.e. lower and upper bounds are both
        specified) then invalid values will be corrected by random samples
        taken between the lower and upper bounds.
        If the interval is semi-open, only one of the bounds being
        specified, then invalid values will be corrected by the value being
        reflected by the same distance from the relevant limit.

        Parameters
        ----------
        val : array-like
            values to examine

        Returns
        -------
        valid : array-like
            values within the support

        Examples
        --------
        >>> b = Interval(0, 10)
        >>> b.valid(11.5)
        8.5
        """
        _val = np.asarray(val, dtype=float)
        if self._closed_bounds:
            valid = np.where(np.isfinite(self.logp(_val)),
                             _val,
                             self.rvs(size=_val.shape))
        else:
            valid = np.where(_val < self._ub, _val, 2 * self._ub - _val)
            valid = np.where(valid > self._lb, valid, 2 * self._lb - valid)

        return valid

    def rvs(self, size=1, random_state=None):
        if self._closed_bounds:
            if random_state is None:
                random_state = self._random_state
            return self.rv.rvs(size=size, random_state=random_state)
        else:
            raise RuntimeError("Can't ask for a random variate from a"
                               " semi-closed interval")
